import math
import struct
import random
import numpy as np
import matplotlib          as mpl
import matplotlib.pyplot   as plt
import matplotlib.patches  as ptc
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

######################################################
#
#    Calculation of multiplicity of cryo EM views 
#    by L. Urzhumtseva & A.Urzhumtsev, May 03, 2024
#    version 5.2
#
#    Urzhumtseva, L., Barchet, Ch., Klaholz, B.P., Urzhumtsev, A.G.
#    Journal of Applied Crystallography, 2024, v. 57, xxx
#    https://doi.org/10.1107/S1600576724002383
# 
######################################################

def InputData(filedata,filemes) : 

    nfdat = open(filedata, 'r')
    nfmes = open(filemes, 'w')

#   default values of the parameters updatable by KeyWords

    FileProj, FileProjW, FileType, FileStruct   = 'unkn', 'unkn',  'unkn', 'unkn'

    GridCls, GridFst, GridType                  =     90,      4,  'mas'
    AngEps, AngMax                              =   0.01,    2.0
    ViewShrink, ViewPert, ViewKappa, ViewRand   =    0.0,    0.0,   0.0 ,      0
    ViewCut, ViewCutLog                         = 'unkn',  'unkn'
    ViewFreqType, ViewWeightType, ViewData      = 'cell',  'uni',  'unkn'

    DrawHist, DrawMap, DrawView                 =  'yes',  'yes',  'yes'
    SizeX, ShiftH, ShowDots                     =    8.0,    0.0,   0.002 
    ShowAxis                                    =    0.0,    0.0 

    ColorHist, ColorPlus, ColorMap, ColorViews  = 'blue',  'red',  'jet' , 'freq'
    BarBottom , BarTop , BarNum , BarInt        = 'unkn', 'unkn', 'unkn',   256
    HistBottom, HistTop, HistNum                = 'unkn', 'unkn', 'unkn'
    HistViewStep                                =    1.0

    TableFreqStep, TableFreqRad, TableType      =    0.0,     [], 'unkn'

    RepRadius, RepAngle, RepLine, RepPhi0       =      3,     12,   0.1,    1.0
    RepColor                                    = 'black'   

    SymCode                                     = 'unkn'
    SymZ                                        =    0.0,    0.0 
    SymX                                        =   90.0,    0.0

#   parameter fixed in the current version
    PrintFlag                                   =      1 

#   technical parameters for drawing in module DrawAll (not updatable by KeyWords) :

#   position of the ColorBar depending on the number of plots chosen (1, 2 or 4)
#   Number of labels in the ColorBar
#   position of the main title
#   parameters of the histogram plot
#   define the label size as a part of the figure size

    ColorBox1   = [0.005, 0.11, 0.01, 0.77]
    ColorBox2   = [0.05 , 0.11, 0.01, 0.77]
    ColorBox4   = [0.05 , 0.53, 0.01, 0.35]
    BoxTitle    = [0.03 , 0.89]
    TitleX      = 0.90
    HistBarWidth , HistBarShift          = 0.4, 0.3
    SizeLabMin, SizeLabMax, SizeLabScale =   6,  24, 1.25

    print('============== input data ==============',file=nfmes)

###############   evential modification of parameters  #########"

    ktype =   0

    while (True) :
      line = nfdat.readline()
      LineCut  = line.rstrip('\n')
      LineCont = LineCut.split()
      kline = len(LineCont)

      if kline == 0 :
         break
      elif LineCont[0][0] == '#' :
         continue
      elif kline <= 2 :
         print('*** Incomplete line in the file of parameters ***')
         print('*** Incomplete line in the file of parameters ***',file=nfmes)
         print(line)
         print(line,file=nfmes)
         exit()
      elif LineCont[1] != '=' :
         print('*** Missed assignement symbol = in the file of parameters ***')
         print('*** Missed assignement symbol = in the file of parameters ***',file=nfmes)
         print(line)
         print(line,file=nfmes)
         exit()

      elif LineCont[0] == 'FileProj'  or LineCont[0] == 'FP':
         if FileProj != 'unkn' :
            print('*** input projections file defined twice - please choose only one ***')
            print('*** input projections file defined twice - please choose only one ***',
                                                                               file=nfmes)
            exit()
         FileProj = LineCont[2]
         print('FileName for the input file with projections    =',file=nfmes)
         print('  ',FileProj,file=nfmes)
         print('FileName for the input file with projections    =')
         print('  ',FileProj)

      elif LineCont[0] == 'FileProjW' or LineCont[0] == 'FW':
         if FileProjW != 'unkn' :
            print('*** output projections file defined twice - please choose only one ***')
            print('*** output projections file defined twice - please choose only one ***',
                                                                                file=nfmes)
            exit()
         FileProjW = LineCont[2]
         print('Output file with projections and weights',5*' ',' =',file=nfmes)
         print('  ',FileProjW,file=nfmes)
         print('Output file with projections and weights',5*' ',' =')
         print('  ',FileProjW)

      elif LineCont[0] == 'FileType'   or LineCont[0] == 'FT':
         FileType = LineCont[2]
         ktype = 1
         print('Imposed FileType for the input projections file = ',FileType,file=nfmes)
      elif LineCont[0] == 'FileStruct' or LineCont[0] == 'FS':
         FileStruct = LineCont[2]
         print('Imposed Keyword to detect projections data      = ',FileStruct,file=nfmes)

      elif LineCont[0] == 'GridCls'  or LineCont[0] == 'GC':
         GridCls = int(LineCont[2])
         print('Updated AngularCercle Number            ',4*' ','  = ',GridCls,file=nfmes)
      elif LineCont[0] == 'GridFst'  or LineCont[0] == 'GF':
         GridFst = int(LineCont[2])
         print('Updated GridFirstCircle Numner          ',4*' ','  = ',GridFst,file=nfmes)
      elif LineCont[0] == 'GridType' or LineCont[0] == 'GT':
         GridType = LineCont[2]
         print('Updated GridType choice                 ',4*' ','  = ',GridType,file=nfmes)

      elif LineCont[0] == 'AngMax'   or LineCont[0] == 'AM':
         AngMax = float(LineCont[2])
         print('Updated Angular radius of a view        ',4*' ','  = ',AngMax,file=nfmes)
      elif LineCont[0] == 'AngEps'   or LineCont[0] == 'AE':
         AngEps = float(LineCont[2])
         print('Updated Contribution Accuracy value     ',4*' ','  = ',AngEps,file=nfmes)

      elif LineCont[0] == 'ViewFreq' or LineCont[0]   == 'VF':
         ViewFreqType = LineCont[2]
         print('Updated ViewsFrequency calculation Method ',3*' ',' = ',ViewFreqType,file=nfmes)
      elif LineCont[0] == 'ViewWeight' or LineCont[0] == 'VW':
         ViewWeightType = LineCont[2]
         print('Updated View-Weight calculation Method    ',3*' ',' = ',ViewWeightType,file=nfmes)
      elif LineCont[0] == 'ViewKappa'  or LineCont[0] == 'VK':
         ViewKappa = float(LineCont[2])
         print('Updated Log-scale parameter K for weight  ',3*' ',' = ',ViewKappa,file=nfmes)
      elif LineCont[0] == 'ViewLog'    or LineCont[0] == 'VL':
         ViewCutLog = float(LineCont[2])
         print('Updated Log2-Cut-frequency value for weight','  ',' = ',ViewCutLog,file=nfmes)
      elif LineCont[0] == 'ViewCut'    or LineCont[0] == 'VC':
         ViewCut = float(LineCont[2])
         print('Updated Cut-frequency value for weight    ',3*' ',' = ',ViewCut,file=nfmes)
      elif LineCont[0] == 'ViewShrink' or LineCont[0] == 'VS':
         ViewShrink = float(LineCont[2])
         print('Updated Shrinking output file par         ',3*' ',' = ',ViewShrink,file=nfmes)
      elif LineCont[0] == 'ViewPert'   or LineCont[0] == 'VP':
         ViewPert = float(LineCont[2])
         print('Updated Perturbation factor for the output',3*' ',' = ',ViewPert,file=nfmes)
      elif LineCont[0] == 'ViewRand'   or LineCont[0] == 'VR':
         ViewRand = int(LineCont[2])
         print('Updated Random seed value                 ',3*' ',' = ',ViewRand,file=nfmes)
      elif LineCont[0] == 'ViewData'   or LineCont[0] == 'VD':
         ViewData = LineCont[2]
         print('Updated Dataformat of output records      ',3*' ',' = ',ViewData,file=nfmes)

      elif LineCont[0] == 'DrawHist' or LineCont[0] == 'DH':
         DrawHist = LineCont[2]
         print('Updated DrawHistogram Flag   ',16*' ',' = ',DrawHist,file=nfmes)
      elif LineCont[0] == 'DrawMap'  or LineCont[0] == 'DM':
         DrawMap  = LineCont[2]
         print('Updated DrawFrequencyMap Flag',16*' ',' = ',DrawMap,file=nfmes)
      elif LineCont[0] == 'DrawView' or LineCont[0] == 'DV':
         DrawView = LineCont[2]
         print('Updated DrawViews Flag       ',16*' ',' = ',DrawView,file=nfmes)

      elif LineCont[0] == 'SizeX'    or LineCont[0] == 'SX':
         SizeX = float(LineCont[2])
         print('Updated X-size of each figure           ',5*' ',' = ',SizeX,file=nfmes)
      elif LineCont[0] == 'ShiftH'   or LineCont[0] == 'SH':
         ShiftH = float(LineCont[2])
         print('Updated Horizontal Shift between figures',5*' ',' = ',ShiftH,file=nfmes)
      elif LineCont[0] == 'ShowDots' or LineCont[0] == 'SD':
         ShowDots = float(LineCont[2])
         print('Updated Dots Radius at the Views plot   ',5*' ',' = ',ShowDots,file=nfmes)
      elif LineCont[0] == 'ShowAxis' or LineCont[0] == 'SA':
         kphi = 3
         if kline >= 5 and (LineCont[3] == ',' or LineCont[3] == ';') :
            LineCont[3] = LineCont[4]
         if kline <= 3 :
            print('*** Incomplete line in the file of parameters ***')
            print('*** Incomplete line in the file of parameters ***',file=nfmes)
            print(line)
            print(line,file=nfmes)
            exit()
         ShowAxis = float(LineCont[2]),float(LineCont[3])
         print('Updated Show direction for the Views plot',4*' ',' = ',ShowAxis,file=nfmes)

      elif LineCont[0] == 'HistNum'    or LineCont[0] == 'HN':
         HistNum = int(LineCont[2])
         print('Updated Histogram Number of Bins',12*' ','  = ',HistNum,file=nfmes)
      elif LineCont[0] == 'HistBottom' or LineCont[0] == 'HB':
         HistBottom = float(LineCont[2])
         print('Updated Histogram Bottom value  ',12*' ','  = ',HistBottom,file=nfmes)
      elif LineCont[0] == 'HistTop'    or LineCont[0] == 'HT':
         HistTop = float(LineCont[2])
         print('Updated Histogram Top value     ',12*' ','  = ',HistTop,file=nfmes)
      elif LineCont[0] == 'HistStep'   or LineCont[0] == 'HS':
         HistViewStep = float(LineCont[2])
         print('Updated Histogram View Step     ',12*' ','  = ',HistViewStep,file=nfmes)

      elif LineCont[0] == 'TableType'    or LineCont[0] == 'TT':
         TableType = LineCont[2]
         print('Updated TableType for stats     ',12*' ','  = ',TableType,file=nfmes)
      elif LineCont[0] == 'TableStep'    or LineCont[0] == 'TS':
         TableFreqStep = float(LineCont[2])
         print('Grid step for the frequency tables',12*' ','= ',TableFreqStep,file=nfmes)
         LineCut2 = LineCut.replace(","," ")
         LineCut3 = LineCut2.replace(";"," ")
         LineCutCont = LineCut3.split()
         mline = len(LineCutCont)
         if mline > 3 :
            for irad in range(3,mline) :
                TableFreqRad.append(float(LineCutCont[irad]))
            print('Radii of regions for the frequency tables :',file=nfmes)
            print('     ',TableFreqRad,file=nfmes)
            
      elif LineCont[0] == 'BarBottom' or LineCont[0] == 'BB':
         BarBottom = float(LineCont[2])
         print('Updated colorBar Bottom value                   = ',BarBottom,file=nfmes)
      elif LineCont[0] == 'BarTop'    or LineCont[0] == 'BT':
         BarTop = float(LineCont[2])
         print('Updated colorBar Top value                      = ',BarTop,file=nfmes)
      elif LineCont[0] == 'BarNum'    or LineCont[0] == 'BN':
         BarNum = int(LineCont[2])
         print('Updated colorBar Number of ticks                = ',BarNum,file=nfmes)
#      elif LineCont[0] == 'BarInt'    or LineCont[0] == 'BI':
#         BarInt = int(LineCont[2])
#         print('Updated colorBar number for Interpolation       = ',BarInt,file=nfmes)

      elif LineCont[0] == 'RepRadius' or LineCont[0] == 'RR':
         RepRadius = int(LineCont[2])
         print('Updated ReperRadius Number     ',13*' ','  = ',RepRadius,file=nfmes)
      elif LineCont[0] == 'RepAngle'  or LineCont[0] == 'RA':
         RepAngle = int(LineCont[2])
         print('Updated ReperAngular Number    ',13*' ','  = ',RepAngle,file=nfmes)
      elif LineCont[0] == 'RepLine'   or LineCont[0] == 'RL':
         RepLine = float(LineCont[2])
         print('Updated ReperLine Width        ',13*' ','  = ',RepLine,file=nfmes)
      elif LineCont[0] == 'RepPhi0'   or LineCont[0] == 'RP':
         RepPhi0 = float(LineCont[2])
         print('Updated ReperLine Width line 0 ',13*' ','  = ',RepPhi0,file=nfmes)
      elif LineCont[0] == 'RepColor'  or LineCont[0] == 'RC':
         RepColor = LineCont[2]
         print('Updated ReperColor             ',13*' ','  = ',RepColor,file=nfmes)

      elif LineCont[0] == 'ColorHist'   or LineCont[0] == 'CH':
         ColorHist = LineCont[2]
         print('Updated Color for the Histogram plot            = ',ColorHist,file=nfmes)
      elif LineCont[0] == 'ColorMap'    or LineCont[0] == 'CM':
         ColorMap = LineCont[2]
         print('Updated Color for the Map of views              = ',ColorMap,file=nfmes)
      elif LineCont[0] == 'ColorViews'  or LineCont[0] == 'CV':
         ColorViews = LineCont[2]
         print('Updated Color for the Views                     = ',ColorViews,file=nfmes)
      elif LineCont[0] == 'ColorPlus'   or LineCont[0] == 'CP':
         ColorPlus = LineCont[2]
         print('Updated Color for the Positive-values Histogram = ',ColorPlus,file=nfmes)

      elif LineCont[0] == 'SymCode' or LineCont[0] == 'YC':
         SymCode = LineCont[2]
         print('Updated Point Symmetry Code  ',16*' ',' = ',SymCode,file=nfmes)
      elif LineCont[0] == 'SymX' or LineCont[0]   == 'YX':
         if kline >= 5 and (LineCont[3] == ',' or LineCont[3] == ';') :
            LineCont[3] = LineCont[4]
         if kline <= 3 :
            print('*** Incomplete line in the file of parameters ***')
            print('*** Incomplete line in the file of parameters ***',file=nfmes)
            print(line)
            print(line,file=nfmes)
            exit()
         SymX = float(LineCont[2]),float(LineCont[3])
         print('Updated X axis to expand views by symmetry    = ',SymX,file=nfmes)
      elif LineCont[0] == 'SymZ' or LineCont[0]   == 'YZ':
         if kline >= 5 and (LineCont[3] == ',' or LineCont[3] == ';') :
            LineCont[3] = LineCont[4]
         if kline <= 3 :
            print('*** Incomplete line in the file of parameters ***')
            print('*** Incomplete line in the file of parameters ***',file=nfmes)
            print(line)
            print(line,file=nfmes)
            exit()
         SymZ = float(LineCont[2]),float(LineCont[3])
         print('Updated Z axis to expand views by symmetry    = ',SymZ,file=nfmes)
#
#     do not modify the default value PrintFlag = 1
#
#      elif LineCont[0] == 'PrintFlag' or LineCont[0] == 'PF':
#         PrintFlag = int(LineCont[2])
#         print('Updated PrintFlag',28*' ',' = ',PrintFlag,file=nfmes)

      else :      
         print('*** Unknown keyword in the file of parameters = ')
         print('*** Unknown keyword in the file of parameters = ',file=nfmes)
         print(LineCont[0])
         print(LineCont[0],file=nfmes)
         exit()

    nfdat.close()
#
#  ----no more data : either end-of-file reached or empty string read ---
#
#   check and, if required, define parameter values
#
    if FileProj == 'unkn' :
       print('*** Input file with projectins is not defined ***')
       print('*** Input file with projectins is not defined ***',file=nfmes)
       exit()

    if FileProjW == FileProj :
       print('*** Input and Output projection files coincide ***')
       print('*** Input and Output projection files coincide ***',file=nfmes)
       exit()

    if PrintFlag < 0 or PrintFlag > 2:
       print('*** Wrong PrintFlag value = ',PrintFlag )
       print('*** Wrong PrintFlag value = ',PrintFlag,file=nfmes)
       exit()

    if GridCls <= 0 :
       print('*** Wrong GridCls number = ',GridCls)
       print('*** Wrong GridCls number = ',GridCls,file=nfmes)
       exit()
    if GridFst <= 1 :
       print('*** Wrong GridFst number = ',GridFst)
       print('*** Wrong GridFst number = ',GridFst,file=nfmes)
       exit()

    if   GridType == 'ros' or GridType == 'ROS' or GridType == 'Ros'  \
                           or GridType == 'R'   or GridType == 'r':
         GridType = 'ros'
    elif GridType == 'mas' or GridType == 'MAS' or GridType == 'Mas'  \
                           or GridType == 'M'   or GridType == 'm':
         GridType = 'mas'
    else :  
         print('*** Wrong GridType number = ',GridType)
         print('*** Wrong GridType number = ',GridType,file=nfmes)
         exit()

    if SizeX <= 0.0 :
       print('*** Wrong Figure X-size, SizeX = ',SizeX)
       print('*** Wrong Figure X-size, SizeX = ',SizeX,file=nfmes)
       exit()
    if ShowDots <= 0.0 or ShowDots > 0.1:
       print('*** Wrong Dots Radius for the Views plot, ShowDots = ',ShowDots)
       print('*** Wrong Dots Radius for the Views plot, ShowDots = ',ShowDots,
                                                                    file=nfmes)
       exit()
    if ShowAxis[1] < 0.0 : ShowAxis[1] += 360.0
    if ShowAxis[0] < 0.0 or ShowAxis[0] > 90.0 or ShowAxis[1] < 0.0 or ShowAxis[0] >= 360.0 :
       print('*** Wrong Direction of projection, ShowAxis = ',ShowAxis)
       print('*** Wrong Direction of projection, ShowAxis = ',ShowAxis,file=nfmes)
       exit()
    if BarInt <= 1 :
       print('*** Wrong colorBar number for Interpolation, BarInt = ',BarInt)
       print('*** Wrong colorBar number for Interpolation, BarInt = ',BarInt,file=nfmes)
       exit()
    if BarNum != 'unkn' and BarNum <= 1 :
       print('*** Wrong colorBar Number of ticks         , BarNum = ',BarNum)
       print('*** Wrong colorBar Number of ticks         , BarNum = ',BarNum,file=nfmes)
       exit()
    if HistNum != 'unkn' and HistNum <= 1 :
       print('*** Wrong Number of Histogram Bars, HistNum = ',HistNum)
       print('*** Wrong Number of Histogram Bars, HistNum = ',HistNum,file=nfmes)
       exit()
    if RepLine < 0.0 :
       print('*** Wrong ReperLine width, RepLine = ',RepLine)
       print('*** Wrong ReperLine width, RepLine = ',RepLine,file=nfmes)
       exit()
    if RepPhi0 < 0.0 :
       print('*** Wrong ReperLine-zero width, RepPhi0 = ',RepPhi0)
       print('*** Wrong ReperLine-zero width, RepPhi0 = ',RepPhi0,file=nfmes)
       exit()
    if RepRadius <  0 :
       print('*** Wrong Number of Reper Circles, RepRadius = ',RepRadius)
       print('*** Wrong Number of Reper Circles, RepRadius = ',RepRadius,file=nfmes)
       exit()
    if RepAngle <  0 :
       print('*** Wrong Number of Reper Meridians, RepAngle = ',RepAngle)
       print('*** Wrong Number of Reper Meridians, RepAngle = ',RepAngle,file=nfmes)
       exit()

    if AngEps <= 0.0 or AngEps >= 1.0 :
       print('*** Wrong Contribution Accuracy, AngEps = ',AngEps )
       print('*** Wrong Contribution Accuracy, AngEps = ',AngEps,file=nfmes)
       exit()
    if AngMax < 0.0 or AngMax > 45.0 :
       print('*** Wrong Angular radius, AngMax = ',AngMax,file=nfmes)
       print('*** Wrong Angular radius, AngMax = ',AngMax)
       exit()

    if AngMax == 0.0 :
       if BarBottom == 'unkn' :
          print('* showing views number, BarBottom assign to be -1')
          print('* showing views number, BarBottom assign to be -1',file=nfmes)
          BarBottom = -1.0
       if HistBottom == 'unkn' :
          print('* showing views number, HistBottom assign to be -1')
          print('* showing views number, HistBottom assign to be -1',file=nfmes)
          HistBottom = -1.0
       print('* Counting integer contributions - AngEps parameter ignored')
       print('* Counting integer contributions - AngEps parameter ignored',file=nfmes)

    if BarTop != 'unkn' and BarBottom != 'unkn' and BarTop <= BarBottom :
       print('*** Wrong colorBar Bottom/Top values = ',BarBottom,BarTop)
       print('*** Wrong colorBar Bottom/Top values = ',BarBottom,BarTop,file=nfmes)
       exit()
    if HistTop != 'unkn' and HistBottom != 'unkn' and HistTop <= HistBottom :
       print('*** Wrong Histogram Bottom/Top bounds = ',HistBottom,HistTop)
       print('*** Wrong Histogram Bottom/Top bounds = ',HistBottom,HistTop,file=nfmes)
       exit()

    if   ColorViews == 'Freq' or ColorViews == 'FREQ' or ColorViews == 'freq' :
         ColorViews = 'freq'
    elif ColorViews == 'Top'  or ColorViews == 'TOP'  or ColorViews == 'top': 
         ColorViews = 'top'
    elif ColorViews == 'Bot'  or ColorViews == 'BOT'  or ColorViews == 'bot': 
         ColorViews = 'bot'
    else :
         print('*** Wrong Color for Views PointPresentation, ColorViews = ',ColorViews)
         print('*** Wrong Color for Views PointPresentation, ColorViews = ',ColorViews,
                                                                            file=nfmes)
         exit()
 
    if   ViewFreqType == 'Cell' or ViewFreqType == 'CELL' or ViewFreqType == 'cell' :
         ViewFreqType = 'cell'
    elif ViewFreqType == 'Num' or ViewFreqType == 'NUM'  or ViewFreqType == 'num': 
         ViewFreqType = 'num'
    elif ViewFreqType == 'Int' or ViewFreqType == 'INT'  or ViewFreqType == 'int': 
         ViewFreqType = 'cell'
         print('* Option Int for ViewsFreqType deleted; ViewsFreq = Cell is used')
         print('* Option Int for ViewsFreqType deleted; ViewsFreq = Cell is used',file=nfmes)
    else :
         print('*** Wrong ViewsFrequency calculation Method = ',ViewFreqType)
         print('*** Wrong ViewsFrequency calculation Method = ',ViewFreqType,file=nfmes)
         exit()
  
    if GridType != 'ros' and ViewFreqType == 'int':
       ViewFreqType = 'cell'
       print('* Non-Rosca grid : ViewsFrequency calculation Method Int substituted by Cell',
                                                                                file=nfmes)
    if GridType == 'mas' :
       if int(GridFst/2) * 2 < GridFst :
          GridFst = GridFst +1
          print('* Masset grid : Grid FirstCircle N intervals increased to un even',GridFst,
                                                                               file=nfmes)

    if FileProjW == 'unkn' :
         if ViewWeightType != 'uni' :
            ViewWeightType  = 'uni'
            print('* Output file with projections is not defined; no weights calculation')
            print('* Output file with projections is not defined; no weights calculation',
                                                                               file=nfmes)

    if   ViewWeightType == 'Unv' or ViewWeightType == 'UNI' or ViewWeightType == 'uni' :
         ViewWeightType  = 'uni'
    elif ViewWeightType == 'Top' or ViewWeightType == 'TOP' or ViewWeightType == 'top' :
         ViewWeightType  = 'top'
    elif ViewWeightType == 'All' or ViewWeightType == 'ALL' or ViewWeightType == 'all': 
         ViewWeightType  = 'all'
    else :
         print('*** Wrong View-Weights Method, ViewWeigth = ',ViewWeightType)
         print('*** Wrong View-Weights Method, ViewWeigth = ',ViewWeightType,file=nfmes)
         exit()

    if   ViewData == 'Stat' or ViewData == 'STAT' or ViewData == 'stat' :
         ViewData  = 'stat'
    elif ViewData == 'Add' or ViewData == 'ADD' or ViewData == 'add' :
         ViewData  = 'add'
    elif ViewData == 'Rep' or ViewData == 'REP' or ViewData == 'rep' :
         ViewData  = 'rep'
    elif ViewData != 'unkn' :
         print('*** Wrong flag to create output file of projections, ViewData = ',ViewData)
         print('*** Wrong flag to create output file of projections, ViewData = ',ViewData,
                                                                                file=nfmes)
         exit()

    if   DrawHist == 'No' or DrawHist == 'NO' or DrawHist == 'no' \
                          or DrawHist == 'N'  or DrawHist == 'n' :
         print('Updated Views Histogram',22*' ',' =  do not show',file=nfmes)
         DrawHist = 'no'
    elif DrawHist == 'Yes' or DrawHist == 'YES' or DrawHist == 'yes' \
                           or DrawHist == 'Y'   or DrawHist == 'y':
         print('Updated Views Histogram',22*' ',' =  show',file=nfmes)
         DrawHist = 'yes'
    else :
         print('*** Wrong flag to plot histograms, DrawHist = ',DrawHist)
         print('*** Wrong flag to plot histograms, DrawHist = ',DrawHist,file=nfmes)
         exit()

    if   DrawMap == 'No' or DrawMap == 'NO' or DrawMap == 'no' \
                         or DrawMap == 'N'  or DrawMap == 'n' :
         print('Updated Frequency Map',24*' ',' =  do not show',file=nfmes)
         DrawMap  = 'no'
    elif DrawMap == 'Yes' or DrawMap == 'YES' or DrawMap == 'yes'\
                          or DrawMap == 'Y'   or DrawMap == 'y':
         print('Updated Frequency Map',24*' ',' =  show',file=nfmes)
         DrawMap  = 'yes'
    else :
         print('*** Wrong flag to plot FrequencyMap, DrawMap = ',DrawMap)
         print('*** Wrong flag to plot FrequencyMap, DrawMap = ',DrawMap,file=nfmes)
         exit()

    if   DrawView == 'No' or DrawView == 'NO' or DrawView == 'no' \
                          or DrawView == 'N'  or DrawView == 'n' :
         print('Updated Views Map',28*' ',' =  do not show',file=nfmes)
         DrawView  = 'no'
    elif DrawView == 'Yes' or DrawView == 'YES' or DrawView == 'yes'\
                           or DrawView == 'Y'   or DrawView == 'y':
         print('Updated Views Map',28*' ',' =  show',file=nfmes)
         DrawView  = 'yes'
    else :
         print('*** Wrong flag to plot point Views, DrawView = ',DrawView)
         print('*** Wrong flag to plot point Views, DrawView = ',DrawView,file=nfmes)
         exit()

    if   TableFreqStep < 0.0 :
         print('*** Wrong value for the radius of checking disks', TableFreqStep)
         print('*** Wrong value for the radius of checking disks ',TableFreqStep,
                                                                          file=nfmes)
         exit()
    elif TableFreqStep > 0.0 and TableType == 'unkn' :
         TableType = 'freq'
         print('* Stat.Table will be calculated for frequencies = ',
                                                            TableType,file=nfmes)

    if TableType != 'unkn' :
       if TableFreqStep == 0.0 : 
          TableFreqStep = 30.0
          print('* Radius of checking disks assigned as',TableFreqStep,file=nfmes)
       if TableFreqRad == [] :
          TableFreqRad.append(TableFreqStep)
          print('* Radius for the checking disks assigned as TableFreqStep =',
                                                            TableFreqStep,file=nfmes)

    print('',file=nfmes)
    if FileProjW == 'unkn' :
         if ViewShrink != 0.0 or ViewPert != 0.0    or ViewKappa  != 0.0 or  \
            ViewRand != 0     or ViewCut  != 'unkn' or ViewCutLog != 'unkn' :
            print('* Output file name is undefined; modification parameters will be ignored')
            print('* Output file name is undefined; modification parameters will be ignored',
                                                                                  file=nfmes)
    else :
         if ViewShrink >= 1.0 and ViewData == 'unkn' :
            FileProjW = 'unkn'
            print('Output file: no new output information required')
            print('Output file: no new output information required',file=nfmes)
            print('* Cancel creating output file') 
            print('* Cancel creating output file',file=nfmes) 

    if GridType == 'mas' :
       if int(GridFst/2) * 2 < GridFst :
          GridFst = GridFst +1
          print('* Masset grid : Grid FirstCircle N intervals increased to un even',GridFst,
                                                                                file=nfmes)
    if ViewPert < 0.0 :
       print('*** Wrong ViewPert value, ViewPert = ',ViewPert)
       print('*** Wrong ViewPert value, ViewPert = ',ViewPert,file=nfmes)
       exit()
    if ViewKappa < 0.0 or ViewKappa > 1.0 :
       print('*** Wrong ViewKappa value, ViewKappa = ',ViewKappa)
       print('*** Wrong ViewKappa value, ViewKappa = ',ViewKappa,file=nfmes)
       exit()

    if ViewCut != 'unkn' :
       if ViewCutLog != 'unkn' :
          print('*** frequency cut-off defined simulateously by both ViewCut and ViewLog')
          print('*** frequency cut-off defined simulateously by both ViewCut and ViewLog',
                                                                               file=nfmes)
          print('*** please choose only one of them')
          print('*** please choose only one of them',file=nfmes)
          exit()
       elif ViewCut <= 0.0 :
          print('*** Wrong ViewCut value, ViewCut = ',ViewCut)
          print('*** Wrong ViewCut value, ViewCut = ',ViewCut,file=nfmes)
          exit()
       else :
          ViewCutLog  = math.log2(ViewCut)
    elif ViewCutLog == 'unkn' :
       ViewCutLog  = 0.0

    if ViewShrink == 0.0 : 
       ViewShrink  = 1.0
    elif ViewShrink < 0.0 or ViewShrink > 1.0 :
       print('*** Wrong ViewShrink value, ViewShrink = ',ViewShrink)
       print('*** Wrong ViewShrink value, ViewShrink = ',ViewShrink,file=nfmes)
       exit()

    SymFamily, SymNumber, SymMirror, SymGroup, SymBase = SymCodeCheck(SymCode,SymX,SymZ)

    if SymCode != 'unkn' :
         if SymGroup :
            print('',file=nfmes)
            print('Point Symmetry Group           ',13*' ','  = ',SymCode,file=nfmes)
            print('      Symmetry Axis Ox (x,y,z) ',13*' ','  = ',SymBase[0],file=nfmes)
            print('      Symmetry Axis Oy (x,y,z) ',13*' ','  = ',SymBase[1],file=nfmes)
            print('      Symmetry Axis Oz (x,y,z) ',13*' ','  = ',SymBase[2],file=nfmes)
         else :
            print('*** Wrong SymmetryGroup Code, SymCode = ',SymCode)
            print('*** Wrong SymmetryGroup Code, SymCode = ',SymCode,file=nfmes)
            exit()

    AngMax   = AngMax * math.pi / 180.
    AngSigma = -0.5*AngMax*AngMax / math.log(AngEps)

#   identify the type of the file of input projections

    if ktype == 0 :
       FileCopy = FileProj.replace("."," ")
       LineCont = FileCopy.split()
       LenName = len(LineCont)
       FileType = LineCont[LenName-1]
       print('Format of the input file of projections identified as',FileType)
       print('Format of the input file of projections identified as',FileType,file=nfmes)

#   conversion cm into inches

    SizeX  = SizeX  / 2.54
    ShiftH = ShiftH / 2.54 

#   define position of the ColorBox Title

    BoxTitleX = ColorBox2[0] - 0.03
    if DrawHist == 'yes' :
       BoxTitleY = ColorBox4[1] + ColorBox4[3] + 0.015
    else :
       BoxTitleY = ColorBox2[1] + ColorBox2[3] + 0.025
    BoxTitle = (BoxTitleX, BoxTitleY)
#
#   ---------- pack program parameters ----------
#
    GridTable    = TableType, TableFreqStep, TableFreqRad
    ColorBox     = (ColorBox1, ColorBox2, ColorBox4, BoxTitle)

    ParamAuxBar  = (BarNum, BarBottom, BarTop, BarInt)
    ParamAuxHist = (HistBarWidth, HistBarShift)
    ParamAuxSize = (SizeLabMin, SizeLabMax, SizeLabScale)

    ParamGrid  = (GridCls, GridFst, GridType, GridTable)
    ParamCalc  = (AngEps, AngMax, AngSigma, ViewFreqType, ViewWeightType, ViewKappa, ViewCutLog)
    ParamFile  = ((PrintFlag, FileProj, FileProjW, FileType), FileStruct)
    ParamStat  = (HistNum, HistBottom, HistTop, HistViewStep)
    ParamColor = (ColorBox, ColorHist, ColorPlus, ColorMap, ColorViews)
    ParamAux   = (ParamAuxBar, TitleX, ParamAuxHist, ParamAuxSize)
    ParamDraw  = (DrawHist, DrawMap, DrawView)
    ParamFig   = (ParamDraw, SizeX, ShiftH, ShowDots)
    ParamReper = (RepRadius, RepAngle, RepLine, RepPhi0, RepColor)
    ParamShow  = (SymCode, SymFamily, SymNumber, SymMirror, SymBase, ShowAxis)
    ParamViews = (ViewShrink, ViewPert, ViewRand, ViewData)

    return ParamGrid, ParamCalc,  ParamStat, ParamFig, ParamColor, ParamFile,         \
           ParamAux,  ParamReper, ParamShow, ParamViews, nfmes

#============================
def SymCodeCheck(SymCode,SymX,SymZ) :

    degrad = math.pi / 180.0 

    SymNumber = 0
    SymMirror = ''
    SymGroup  = True
    SymXvec   = [1., 0., 0.]
    SymYvec   = [0., 1., 0.]
    SymZvec   = [0., 0., 1.]

    SymFamily = SymCode[0]
    if SymFamily == 'c' :
       SymFamily == 'C'
    if SymFamily == 'd' :
       SymFamily == 'D'
    if SymFamily == 's' :
       SymFamily == 'S'
    if SymFamily == 't' :
       SymFamily == 'T'
    if SymFamily == 'o' :
       SymFamily == 'O'
    if SymFamily == 'i' :
       SymFamily == 'I'

    nchar = len(SymCode)

    if SymCode == 'Ci' or SymCode == 'ci':
       SymCode =  'S2'   

    if SymCode == 'Cs' or SymCode == 'C1v' or SymCode == 'cs' or SymCode == 'c1v' :
       SymCode =  'C1h'   

    if SymCode == 'unkn' :
       SymBase = [SymXvec, SymYvec, SymZvec]
       return SymFamily, SymNumber, SymMirror, SymGroup, SymBase

#   polyhedral groups (no numeric component)

    elif SymFamily == 'T' or SymFamily == 'O' or SymFamily == 'I' :

       if nchar == 1 :
          pass

       elif nchar == 2 :
          SymMirror = SymCode[1]
          if SymCode != 'Td' and SymCode != 'Th' and SymCode != 'Oh' and SymCode != 'Ih' : 
             SymGroup  = False
          
       else :
          SymGroup  = False

#   axial groups (number defines the order of the principal rotation axis)

    elif SymFamily == 'C' or SymFamily == 'S' or SymFamily == 'D' :

#      actually no number defined

       if nchar == 1 :
          SymGroup  = False

       else :

#         extract mirror code is present 

          SymMirror = SymCode[nchar-1]

          if (SymFamily == 'C' and (SymMirror == 'h' or SymMirror == 'v') ) or \
             (SymFamily == 'D' and (SymMirror == 'h' or SymMirror == 'd') ) :   
             nchar = nchar - 1
          else :
             SymMirror = ''     

#         extract the axis order

          code_axis = SymCode[1:nchar] 
          if code_axis.isnumeric() :
             SymNumber = int(code_axis)
          else :
             SymGroup = False  

#   unconventional code for the point group

    else :
       SymGroup  = False

#   orthogonalize Ox and Oz if symmetries used

    thZ , phiZ = SymZ[0] * degrad , SymZ[1] * degrad 
    thX , phiX = SymX[0] * degrad , SymX[1] * degrad 
    Z_x, Z_y, Z_z = math.cos(phiZ)*math.sin(thZ), math.sin(phiZ)*math.sin(thZ), math.cos(thZ)
    X_x, X_y, X_z = math.cos(phiX)*math.sin(thX), math.sin(phiX)*math.sin(thX), math.cos(thX)

    dot_prod = X_x * Z_x + X_y * Z_y + X_z * Z_z
    X_x      = X_x - dot_prod * Z_x
    X_y      = X_y - dot_prod * Z_y
    X_z      = X_z - dot_prod * Z_z

#   rotation matrix, in columns, is composed of the coordinates of the main symmetry axes

    Xnorm = math.sqrt(X_x * X_x + X_y * X_y + X_z * X_z)
    SymXvec = [X_x/Xnorm, X_y/Xnorm, X_z/Xnorm]
    SymZvec = [Z_x      , Z_y      , Z_z      ]
    SymYvec = [Z_y * X_z - Z_z * X_y , Z_z * X_x - Z_x * X_z , Z_x * X_y - Z_y * X_x]

    SymBase = [SymXvec, SymYvec, SymZvec]

    return SymFamily, SymNumber, SymMirror, SymGroup, SymBase

#============================
def GetViews(ParamFile,nfmes) : 

#   all intermediate rotation matrices are given by columns, i.e., the elements are
#   R[0] = R[0,0], R[1] = R[1,0], R[2] = R[2,0], R[3] = R[0,1], R[4] = R[1,1], ...

    (PrintFlag,FileProj,FileProjW,FileType),FileStruct = ParamFile

    if   FileType == 'test' :
         projections,FileStructN = InputProjectionsTest(FileProj,FileStruct,nfmes)
         views_inp               = ProjectionsViewsRelion(projections,nfmes)

    elif FileType == 'star' :
         projections,FileStructN = InputProjectionsStar(FileProj,FileStruct,nfmes)
         views_inp               = ProjectionsViewsRelion(projections,nfmes)

    elif FileType == 'hed' :
         projections,FileStructN = InputProjectionsImagic(FileProj,FileStruct,nfmes)
         views_inp               = ProjectionsViewsImagic(projections,nfmes)

    else :
       print('UNKNOWN FILE TYPE FOR PROJECTIONS',file=nfmes)
       print('UNKNOWN FILE TYPE FOR PROJECTIONS')
       exit ()

    nprojs = len(projections)

    print('',file=nfmes)
    print('============== end of input data ==============',file=nfmes)
    print('',file=nfmes)

    print(f'Total number of input views',' '*14,f'{nprojs:12}')
    print(f'Total number of input views',' '*14,f'{nprojs:12}', file=nfmes)

    ParamFileN = ((PrintFlag,FileProj,FileProjW,FileType),FileStructN)

    return views_inp, ParamFileN

#============================
def InputProjectionsTest(FileProj,FileStruct,nfmes) : 

    projections = []

    nfproj = open(FileProj, 'r')

    while True:
      line = nfproj.readline()
      if not line :
         break
      LineCont = line.split()
      projections.append([LineCont[0],LineCont[1],LineCont[2]])

    nfproj.close()

    FileStructN = 0

    return projections,FileStructN
   
#============================
def InputProjectionsStar(FileProj,FileStruct,nfmes): 

    projections = []

#    NumPosAngles = positions of psi, rot, tilt in the respective lines
#    NumLinAngles = numbers of lines with the descriptions of psi, rot, tilt

    nfproj = open(FileProj, 'r')
# 
#   block name is defined
# 

    if FileStruct != 'unkn' :

       print ('*** Input projections file : Search for the keyword ',FileStruct,
              file=nfmes)

       NumTitles = FindKey1(nfproj,FileStruct)
       if NumTitles < 0 :
          print ('*** EOF reached before the data block found',file=nfmes)
          print ('*** EOF reached before the data block found')
          exit()

#      find the following 'loop_' record

       NumTitle = FindKey2(nfproj,'loop_')
       if NumTitle < 0 :
          print ('*** the keyword is not followed by loop_')
          print ('*** the keyword is not followed by loop_',file=nfmes)
          exit()
       else :
          NumTitles = NumTitles + NumTitle

#      read label records

       NumTitle,MaxLabel,NumPosAngles,NumLinAngles,line = GetLabels(nfproj)
       if NumTitle < 0:
          print ('*** wrong label descriptors')
          print ('*** wrong label descriptors',file=nfmes)
          exit()

       NlinePsi  = NumLinAngles[0] + NumTitles - 1
       NlineRot  = NumLinAngles[1] + NumTitles - 1
       NlineTilt = NumLinAngles[2] + NumTitles - 1
       NumLinAngles = (NlinePsi,NlineRot,NlineTilt)

       NumTitles = NumTitles + NumTitle
       if NumPosAngles[0] * NumPosAngles[1] * NumPosAngles[2] == 0 :
          print ('*** missed angles labels in the required data block ')
          print ('*** missed angles labels in the required data block ',file=nfmes)
          exit()

#      read data

       projections,NumLines,LastLine = GetDataProj(nfproj,MaxLabel,NumPosAngles,line)
       if NumLines < 0 :
          print ('*** wrong record in the list of projections')
          print ('*** wrong record in the list of projections',file=nfmes)
          print (LastLine)
          print (LastLine,file=nfmes)
          exit()
# 
#   block name is not defined
#
    else :

       print ('Input projections file : Search for the data block with angle labels',
              file=nfmes)

       NumTitles = 0

       while (True) :

#         read all data till 'loop_'

          NumTitle = FindKey1(nfproj,'loop_')

          if NumTitle < 0 :
             print ('*** EOF reached')
             print ('*** EOF reached',file=nfmes)
             exit()

          NumTitles = NumTitles + NumTitle
  
#         read label records

          NumTitle,MaxLabel,NumPosAngles,NumLinAngles,line = GetLabels(nfproj)
          if NumTitle < 0 :
             print ('*** wrong label descriptors or EOF reached')
             print ('*** wrong label descriptors or EOF reached',file=nfmes)
             print (line)
             print (line,file=nfmes)
             exit()

          NlinePsi  = NumLinAngles[0] + NumTitles - 1
          NlineRot  = NumLinAngles[1] + NumTitles - 1
          NlineTilt = NumLinAngles[2] + NumTitles - 1
          NumLinAngles = (NlinePsi,NlineRot,NlineTilt)

          NumTitles = NumTitles + NumTitle

          if NumPosAngles[0] >= 0 and NumPosAngles[1] >= 0 and NumPosAngles[2] >= 0 :
             break
          else :
             NumTitles = NumTitles + 1   

       if NumPosAngles[0] < 0 or NumPosAngles[1] < 0 or NumPosAngles[2] < 0 :
          print ('*** failed to find a data block with angles')
          print ('*** failed to find a data block with angles',file=nfmes)
          exit()

#      the data block with projections found; read data
 
       projections,NumLines,LastLine = GetDataProj(nfproj,MaxLabel,NumPosAngles,line)

       if NumLines < 0 :
          print ('*** wrong record in the list of projections')
          print ('*** wrong record in the list of projections',file=nfmes)
          print (LastLine)
          print (LastLine,file=nfmes)
          exit()
#
#   all information extracted
#
    nfproj.close()

    FileStructN = (NumTitles,MaxLabel,NumPosAngles,NumLinAngles)

    return projections,FileStructN     
                         
#============================
def InputProjectionsImagic(FileProj,FileStruct,nfmes) : 

    projections = []

    nfproj = open(FileProj, 'rb')

#   detailed Imagic format (not required here)
#    form = '@14i 4s 2i 12f 80s f 14i 3f 31i 3f 2i 3f i 13f 2i 77f 228s'

#   format to extract angles, number of blocks and the machine stamp
    form = '@12s i 240s 3f 4s i 748s'

    nbytes = 1024

    while True :
      line = nfproj.read(nbytes)
      if not line :
         break
      LineCont = struct.unpack(form, line)

      Mstamp = LineCont[7]

      if Mstamp != 33686018 :
         print('input file has been created in a non-compatible computer')
         print('input file has been created in a non-compatible computer',file=nfmes)
         print('machine stamp is ',Mstamp,' (see imagic_format.pdf)')
         print('machine stamp is ',Mstamp,' (see imagic_format.pdf)',file=nfmes)
         exit()

      projections.append([LineCont[3],LineCont[4],LineCont[5]])

      Nblocks = LineCont[1]

      if Nblocks > 1 :
         for ib in range(Nblocks-1) :
             line = nfproj.read(nbytes)

    nfproj.close()

    FileStructN = (0,0,0)

    return projections,FileStructN

#============================
def FindKey1(nfproj,KeyWord) :

    NumTitle = 0
    while (True) :
       line = nfproj.readline()
       if not line :
          NumTitle = -1
          break

       NumTitle = NumTitle + 1
       LineCont = line.split()
       kline = len(LineCont)
       if kline > 0 and LineCont[0] == KeyWord :
          break
    
    return NumTitle
                         
#============================
def FindKey2(nfproj,KeyWord) :

    NumTitle = 0
    while (True) :
       line = nfproj.readline()
       if not line :
          NumTitle = -1
          break
       NumTitle = NumTitle + 1
       LineCont = line.split()
       kline = len(LineCont)
       if kline > 0 :
          if LineCont[0][0] == '#' or LineCont[0][0] == ';' :
             continue
          elif LineCont[0] == KeyWord :
             break
          else :
             NumTitle = -1
             break

    return NumTitle
                         
#============================
def GetLabels(nfproj) :

    NumTitle , MaxLabel              = 0, 0
    NumPosPsi, NumPosRot, NumPosTilt = -1, -1, -1
    NumLinPsi, NumLinRot, NumLinTilt = -1, -1, -1

    while (True) :
      line = nfproj.readline()
      if not line :
         NumTitle = -1
         break

      LineCont = line.split()
      kline = len(LineCont)

      if kline == 0 :
         NumTitle = -1
         break

      NumTitle = NumTitle + 1
      if LineCont[0][0] == '#' or LineCont[0][0] == ';':
         continue
      elif kline <= 1 :
         NumTitle = -1
         break
      elif LineCont[0][0] != '_' :
         NumTitle = NumTitle - 1
         break
      else :
         mline = len(LineCont[0])
         if mline < 5 or LineCont[0][1] != 'r' or LineCont[0][2] != 'l' or LineCont[0][3] != 'n':
            NumTitle = -1
            break
         NumPos = int(LineCont[1].strip('#'))
         if NumPos > MaxLabel :
            MaxLabel = NumPos
         if   LineCont[0] == '_rlnAnglePsi'  :
              NumPosPsi  = NumPos - 1
              NumLinPsi  = NumTitle
         elif LineCont[0] == '_rlnAngleRot'  :
              NumPosRot  = NumPos - 1
              NumLinRot  = NumTitle
         elif LineCont[0] == '_rlnAngleTilt' :
              NumPosTilt = NumPos - 1
              NumLinTilt = NumTitle
         else :
              continue

    NumPosAngles = (NumPosPsi, NumPosRot, NumPosTilt)
    NumLinAngles = (NumLinPsi, NumLinRot, NumLinTilt)

    return NumTitle, MaxLabel, NumPosAngles, NumLinAngles, line
                       
#============================
def GetDataProj(nfproj,MaxLabel,NumPosAngles,line) :

    projections = []

    NumPosPsi, NumPosRot, NumPosTilt = NumPosAngles
    NumLines = 0

#   first projections already read; process it and read next projections

    while True :

       LineCont = line.split()
       kline = len(LineCont)

       if   kline == 0 :
            break
       elif LineCont[0][0] == '#' or LineCont[0][0] == ';' :
            NumLines = NumLines + 1

#      line may end up by some extra symbols like '/', '\' etc making kline > MaxLabel

       elif kline >= MaxLabel :
            projections.append([LineCont[NumPosPsi],LineCont[NumPosRot],LineCont[NumPosTilt]])
            NumLines = NumLines + 1
       else :
            NumLines = -1
            return projections,NumLines,line

#      next projection

       line = nfproj.readline()
       if not line :
          line = 'EOF'
          break

    return projections, NumLines, line
                         
#============================
def ProjectionsViewsRelion(projections,nfmes) : 

    Nviews = len(projections)

    views = [ [0, 0, 0] for i in range(Nviews)]

    degrad = math.pi/180.

    for iv in range(Nviews):
       psi  = float(projections[iv][0]) * degrad
       rot  = float(projections[iv][1]) * degrad
       tilt = float(projections[iv][2]) * degrad

#      matrix of the rotation operator of the object is RotZ(-psi)RotY(-tilt)RotZ(-phi)
#      matrix of the inverse operator to align a projection with the object is
#      RotZ(phi)RotY(tilt)RotZ(psi) with the last column as the "views"

       RotMatrix = RotTotal(RotZ,rot,RotY,tilt,RotZ,psi)

       xv,yv,zv = (RotMatrix[6],RotMatrix[7],RotMatrix[8])

       if zv >= 0.0 :
          views[iv] = [xv,yv,zv]
       else :
          views[iv] = [-xv,-yv,-zv]

    return views         
                         
#============================
def ProjectionsViewsImagic(projections,nfmes) : 

    Nviews = len(projections)

    views = [ [0, 0, 0] for i in range(Nviews)]

    degrad = math.pi/180.

    for iv in range(Nviews):
       alpha = float(projections[iv][0]) * degrad
       beta  = float(projections[iv][1]) * degrad
       gamma = float(projections[iv][2]) * degrad

#      matrix of the rotation operator of the object is RotZ(-alpha)RotX(-beta)RotZ(-gamma)
#      matrix of the inverse operator to align a projection with the object is
#      RotZ(gamma)RotX(beta)RotZ(alpha) with the last column as the "views"

       RotMatrix = RotTotal(RotZ,gamma,RotX,beta,RotZ,alpha)

       xv,yv,zv = (RotMatrix[6],RotMatrix[7],RotMatrix[8])

       if zv >= 0.0 :
          views[iv] = [xv,yv,zv]
       else :
          views[iv] = [-xv,-yv,-zv]

    return views     
                             
#============================
def RotTotal(Rot1,kappa1,Rot2,kappa2,Rot3,kappa3) : 

    rotmat1, rotmat2, rotmat3 = Rot1(kappa1), Rot2(kappa2), Rot3(kappa3)

    rotmat23  = MultMat(rotmat2,rotmat3)
    RotMatrix = MultMat(rotmat1,rotmat23)

    return RotMatrix     
                             
#============================
def RotX(kappa) : 

    rotation = [0 for j in range(9)] 

    ck, sk = math.cos(kappa), math.sin(kappa)

    rotation[0] , rotation[3] , rotation[6] = 1.0 , 0.0 , 0.0
    rotation[1] , rotation[4] , rotation[7] = 0.0 ,  ck , -sk
    rotation[2] , rotation[5] , rotation[8] = 0.0 ,  sk ,  ck 

    return rotation     
                             
#============================
def RotY(kappa) : 

    rotation = [0 for j in range(9)] 

    ck, sk = math.cos(kappa), math.sin(kappa)

    rotation[0] , rotation[3] , rotation[6] =  ck , 0.0 ,  sk
    rotation[1] , rotation[4] , rotation[7] = 0.0 , 1.0 , 0.0
    rotation[2] , rotation[5] , rotation[8] = -sk , 0.0 ,  ck 

    return rotation     
                             
#============================
def RotZ(kappa) : 

    rotation = [0 for j in range(9)] 

    ck, sk = math.cos(kappa), math.sin(kappa)

    rotation[0] , rotation[3] , rotation[6] =  ck , -sk , 0.0
    rotation[1] , rotation[4] , rotation[7] =  sk ,  ck , 0.0
    rotation[2] , rotation[5] , rotation[8] = 0.0 , 0.0 , 1.0 

    return rotation     
                              
#============================
def MatVect(rot,x,y,z) : 

    RotVect = [0 for i in range(3)]

    RotVect[0] = rot[0]*x + rot[3]*y + rot[6]*z
    RotVect[1] = rot[1]*x + rot[4]*y + rot[7]*z
    RotVect[2] = rot[2]*x + rot[5]*y + rot[8]*z

    return RotVect
                                                                                  
#============================
def MultMat(rot1,rot2) : 

#   matrix product rot1*rot2 (matrices are kept by column)

    rot3 = [0 for i in range(9)]

    rot3[0] = rot1[0]*rot2[0] + rot1[3]*rot2[1] + rot1[6]*rot2[2]
    rot3[1] = rot1[1]*rot2[0] + rot1[4]*rot2[1] + rot1[7]*rot2[2]
    rot3[2] = rot1[2]*rot2[0] + rot1[5]*rot2[1] + rot1[8]*rot2[2]
    rot3[3] = rot1[0]*rot2[3] + rot1[3]*rot2[4] + rot1[6]*rot2[5]
    rot3[4] = rot1[1]*rot2[3] + rot1[4]*rot2[4] + rot1[7]*rot2[5]
    rot3[5] = rot1[2]*rot2[3] + rot1[5]*rot2[4] + rot1[8]*rot2[5]
    rot3[6] = rot1[0]*rot2[6] + rot1[3]*rot2[7] + rot1[6]*rot2[8]
    rot3[7] = rot1[1]*rot2[6] + rot1[4]*rot2[7] + rot1[7]*rot2[8]
    rot3[8] = rot1[2]*rot2[6] + rot1[5]*rot2[7] + rot1[8]*rot2[8]

    return rot3
                                                         
#============================
def SymViews(views_inp,ParamShow,nfmes) :

#   generate symmetric views if required

    SymCode = ParamShow[0]

    if SymCode == 'unkn' :
       views = views_inp
       Nsym = 1
       print('   No symmetries used')

    else :
       SymMatrices = GetSymMatrices(ParamShow)
       SymMatrices = AxesSymMatrices(SymMatrices,ParamShow)
       views       = GetSymViews(views_inp,SymMatrices)

       Nsym = len(SymMatrices)
       nprojs = len(views)

       print(f'Number of point symmetry operators  ',' '*5,f'{Nsym:12}', file=nfmes)
       print(f'Total number of generated views     ',' '*5,f'{nprojs:12}', file=nfmes)

       print(nprojs,' views in total after applying ',Nsym,' symmetry operators')

    return views

#============================
def GetSymMatrices(ParamShow) : 

    SymCode,SymFamily,SymNumber,SymMirror,SymBase,ShowAxis = ParamShow

    SymMatrices  = []

#   First, get matrices in the standard orientation of the symmetry axes : 
#      SymZvec coincides with the projection axis (spherical angles 0, 0) 
#      SymXvec is horizontal, from left to right (spherical angles 90, 0)

#   first operation is a trivial symmetry, C1

    SymMatrices.append(RotZ(0.0))
    Nsym = 1

#   principal rotation axis aligned with Oz
#   diagonal icosaedre 5-fold axis is aligned with Oz

    NumberSymMain = SymNumber

    if SymFamily == 'I' : 
       NumberSymMain, invert = 5 , 0
    elif SymFamily == 'O' :
       NumberSymMain, invert = 4 , 0
    elif SymFamily == 'T' :
       if   SymMirror == 'h' :
            NumberSymMain, invert = 4 , 0
       elif SymMirror == 'd' :
            NumberSymMain, invert = 4 , 1
       else :
            NumberSymMain, invert = 4 , 0

    if NumberSymMain > 1 :
       delphi = math.pi * 2.0 / float(NumberSymMain)

       if SymFamily == 'S' or (SymFamily == 'D' and SymMirror == 'd') :
          invert = 1
       else :
          invert = 0

       for isym in range(1,NumberSymMain) :
            RotMat = RotZ(delphi * isym)
            if invert == 1 : 
               RotMat[8] = -1.0
               invert = -invert
            elif invert == -1 :
               invert = -invert
            SymMatrices.append(RotMat)

#   extra rotation axis (perpendicular to Oz)

    if SymFamily == 'D' or SymFamily == 'O' or SymFamily == 'T' : 
       RotMat = RotX(math.pi)
       if SymCode == 'Td' :
          RotMat[0] = -1.0

       Nsym = len(SymMatrices)
       for isym in range(Nsym) : 
           SymMatrices.append(MultMat(RotMat,SymMatrices[isym]))

#   diagonal cube axes 

    if SymFamily == 'T' or SymFamily == 'O' : 
       for ir in range(9) :
           RotMat[ir] = 0.0
       RotMat[1], RotMat[4], RotMat[7] = 1.0, 1.0, 1.0

       Nsym = len(SymMatrices)
       for isym in range(Nsym) : 
           SymMatrices.append(MultMat(RotMat,SymMatrices[isym]))

#   icosaedre 3-fold and 2-fold axes are in the xOz plan

    if SymFamily == 'I' : 

#      3-fold

       norm5   = math.sqrt((5.0 - math.sqrt(5.0)) / 2.0)
       norm3   = math.sqrt((3.0 + math.sqrt(5.0)) / 6.0)
       phi53   = math.arcos(norm3 / norm5)
       RotMat1 = RotY(-phi53)
       RotMat2 = RotZ(2.0 * math.pi / 3.0)
       RotMat3 = MultMat(RotMat2,RotMat1)
       RotMat1 = RotY(phi53)
       RotMat  = MultMat(RotMat1,RotMat3)

       Nsym = len(SymMatrices)
       for isym in range(Nsym) : 
           SymMatrices.append(MultMat(RotMat,SymMatrices[isym]))

#      2-fold in the plan x0z

       phi52   = math.arcos( math.sqrt ((5.0 - math.sqrt(5.0)) / 10.0) )
       RotMat1 = RotY(-phi52)
       RotMat2 = RotZ(math.pi)
       RotMat3 = MultMat(RotMat2,RotMat1)
       RotMat1 = RotY(phi52)
       RotMat  = MultMat(RotMat1,RotMat3)

       Nsym = len(SymMatrices)
       for isym in range(Nsym) : 
           SymMatrices.append(MultMat(RotMat,SymMatrices[isym]))

#      orthogonal 2-fold aligned with Oy

       RotMat  = RotY(math.pi)

       Nsym = len(SymMatrices)
       for isym in range(Nsym) : 
           SymMatrices.append(MultMat(RotMat,SymMatrices[isym]))
 
#   mirror plane (perpendicular to Oz)

    if SymMirror == 'h' :
       for ir in range(9) :
           RotMat[ir] = 0.0
       RotMat[0], RotMat[3], RotMat[8] = 1.0, 1.0, -1.0

       Nsym = len(SymMatrices)
       for isym in range(Nsym) : 
           SymMatrices.append(MultMat(RotMat,SymMatrices[isym]))

#   mirror plane (containing Oz)

    if SymMirror == 'v' :

       phi = math.pi * 2.0 / float(NumberSymMain)

       RotMat = RotZ(phi)
       RotMat[3], RotMat[4] = -RotMat[3], -RotMat[4]

       Nsym = len(SymMatrices)
       for isym in range(Nsym) : 
           SymMatrices.append(MultMat(RotMat,SymMatrices[isym]))

    return SymMatrices
                             
#============================
def AxesSymMatrices (SymMatrices, ParamShow) :

#   now take into account the orientation of the symmetry axes
#   with respect axes of the Lambert projection

    RotBase  = [0 for i in range(9)]
    RotBaseI = [0 for i in range(9)]

    SymCode,SymFamily,SymNumber,SymMirror,SymBase,ShowAxis = ParamShow

#   no modification required; symmetry axes coincide with the projection axes

    if SymBase[0] == [1.,0.,0.] and SymBase[1] == [0.,1.,0.] and SymBase[2] == [0.,0.,1.] :
       return SymMatrices

#   modifying symmery matrices
#   equivalence of two oppsite views results in addition of the center of inversion

    else : 

       RotBase[0] , RotBase[1] , RotBase[2]  = SymBase[0]
       RotBase[3] , RotBase[4] , RotBase[5]  = SymBase[1]
       RotBase[6] , RotBase[7] , RotBase[8]  = SymBase[2]

       RotBaseI[0], RotBaseI[1], RotBaseI[2] = SymBase[0][0], SymBase[1][0], SymBase[2][0]
       RotBaseI[3], RotBaseI[4], RotBaseI[5] = SymBase[0][1], SymBase[1][1], SymBase[2][1]
       RotBaseI[6], RotBaseI[7], RotBaseI[8] = SymBase[0][2], SymBase[1][2], SymBase[2][2]

       Nsym = len(SymMatrices)
       for isym in range(Nsym) : 

           RotCoord          = MultMat(SymMatrices[isym],RotBaseI)
           SymMatrices[isym] = MultMat(RotBase,RotCoord)

    return SymMatrices     

#============================
def GetSymViews(views_inp,SymMatrices) : 
    
    Nsym = len(SymMatrices)
    nviews = len(views_inp)
    views  = []

    for isym in range(Nsym) :
        for iview in range(nviews) :
            vi = views_inp[iview]
            xv,yv,zv = MatVect(SymMatrices[isym],vi[0],vi[1],vi[2])

            if zv >= 0.0:
               views.append([xv,yv,zv])
            else:
               views.append([-xv,-yv,-zv])
    
    return views
                                                      
#============================
def ViewsToShow(views,ParamShow) :

#   reorient the views to show if required

    SymCode,SymFamily,SymNumber,SymMirror,SymBase,ShowAxis = ParamShow

    degrad = math.pi / 180.

#   additional rotation to put the view axis normal to the page

    viewtheta, viewphi = ShowAxis

    if viewtheta == 0.0 and viewphi == 0.0 :
       print('   Projection axis not changed')

    else :
       viewtheta = viewtheta * degrad
       viewphi   = viewphi   * degrad
       RotView = RotTotal(RotZ,viewphi,RotY,-viewtheta,RotZ,-viewphi)

       nviews = len(views)

       for iv in range(nviews) :

          vinit = views[iv]

          RotVect  = MatVect(RotView,vinit[0],vinit[1],vinit[2])
          xv,yv,zv = RotVect

          if zv >= 0.0 :
             views[iv] = [xv,yv,zv]
          else :
             views[iv] = [-xv,-yv,-zv]

       print('   Views were reoriented accorging to ShowAxis')
   
    return views

#============================
def GridCenters (ParamGrid,nfmes) : 

#   calculate coordinates of the grid cells on a unit sphere

    GridCls, GridFst, GridType, GridTable = ParamGrid

    if GridType == 'ros' :     
       CellsXYZ = GridCentersRosca (GridCls,nfmes)
       print('Rosca grid cell is used',file=nfmes)     

    elif GridType == 'mas' :      
       CellsXYZ = GridCentersMasset (GridCls,GridFst,nfmes)
       print('Masset grid cell is used',file=nfmes)     

    else :
       print('*** Unknown type of the grid cell for the grid coordinates')
       print('*** Unknown type of the grid cell for the grid coordinates',file=nfmes)
       exit()

    return CellsXYZ
                                                                                                                                                                                                                          
#============================
def GridCentersRosca (GridCls,nfmes) : 

    CellsXYZ = [ [ [0.0, 0.0, 0.0] for j in range(GridCls*8+1)] for i in range(GridCls+1)]

#   layers are counted from mg = 1 (for the inner layer)

#   Uxyz2, Vxyz2 - coordinates of the grid points of the outer circle of the layer;
#   Uxyz1, Vxyz1 - coordinates of the grid points of the inner circle of the layer;
#   Middle of the shortest cell diagonal is taken as the cell center.
#       Diagonal grid cells [mg][mg] and [mg][mg+1] have the same centre CellsXYZ   
#   since they are merged in a single sector-cells ; the same for other diagonals

    Uxyz1 = [0.0 for i in range(GridCls+1)]
    Vxyz1 = [0.0 for i in range(GridCls+1)]

    Uxyz2 = [0.0 for i in range(GridCls+1)]
    Vxyz2 = [0.0 for i in range(GridCls+1)] 

    for mg in range(1,GridCls+1):

        m2,  m4,  m6  = mg * 2, mg * 4, mg * 6
        m21, m41, m61 = m2 + 1, m4 + 1, m6 + 1       
        m81           = mg * 8 +1

        Rm = math.sqrt(2.) * float(mg) / float(GridCls)
        for kg in range(mg+1):
            phi       = kg * math.pi / (4. * float(mg))
            Uxyz2[kg], Vxyz2[kg] = Rm * math.cos(phi), Rm * math.sin(phi)

        for k1 in range(mg):
            k2 = k1 + 1
            Xmk, Ymk = (Uxyz1[k1] + Uxyz2[k2])/2. , (Vxyz1[k1] + Vxyz2[k2])/2.
            Rmk2     = (Xmk*Xmk + Ymk*Ymk) / 2.
            scalexy  = math.sqrt(1. - Rmk2/2.)
            x_sphere , y_sphere  = Xmk * scalexy, Ymk * scalexy
            z_sphere = 1.0 - Rmk2

            CellsXYZ[mg][ k2]     = [ x_sphere,  y_sphere, z_sphere]
            CellsXYZ[mg][ k2+m2]  = [-y_sphere,  x_sphere, z_sphere]
            CellsXYZ[mg][ k2+m4]  = [-x_sphere, -y_sphere, z_sphere]
            CellsXYZ[mg][ k2+m6]  = [ y_sphere, -x_sphere, z_sphere]
            CellsXYZ[mg][-k2+m21] = [ y_sphere,  x_sphere, z_sphere]
            CellsXYZ[mg][-k2+m41] = [-x_sphere,  y_sphere, z_sphere]
            CellsXYZ[mg][-k2+m61] = [-y_sphere, -x_sphere, z_sphere]
            CellsXYZ[mg][-k2+m81] = [ x_sphere, -y_sphere, z_sphere]

        for kg in range(mg+1):
            Uxyz1[kg], Vxyz1[kg] = Uxyz2[kg], Vxyz2[kg]

    return CellsXYZ
                                                                                                            
#============================
def GridCentersMasset (GridCls,GridFst,nfmes) : 

    CellsXYZ = [ [ [0.0, 0.0, 0.0] for j in range(GridFst * (2*i+1) + 1)] for i in range(GridCls+1)]

#   layers are counted from mg = 1 (for the inner layer)
#   Nm is the number of cells in the layer mg
#   Rm is the radius of the circle crossing the middle of the cell

    for mg in range(1, GridCls+1):

        Rm = math.sqrt(2.) * float(mg-0.5) / float(GridCls)
        Nm = (2 * mg - 1)  * GridFst 
        dphi = 2.0 * math.pi / float(Nm)

        for kg in range(1, Nm+1):
            phi      = float(kg - 0.5) * dphi
            Xmk, Ymk = Rm * math.cos(phi) , Rm * math.sin(phi)

            Rmk2     = Rm * Rm / 2.
            scalexy  = math.sqrt(1. - Rmk2/2.)
            x_sphere , y_sphere  = Xmk * scalexy, Ymk * scalexy
            z_sphere = 1.0 - Rmk2

            CellsXYZ[mg][kg] = [ x_sphere,  y_sphere, z_sphere]

    return CellsXYZ
                                                         
#============================
def ViewsParameters (ParamGrid,views,nfmes) : 

#   radv, angv are polar coordinates of the view's projection into disk
#   mv, kv are indices of the respective grid cell

    GridCls, GridFst, GridType, GridTable = ParamGrid

    Nviews = len(views)

    viewsR = [0 for i in range(Nviews)]
    viewsA = [0 for i in range(Nviews)]
    viewsM = [0 for i in range(Nviews)]
    viewsK = [0 for i in range(Nviews)]

    pi2  = math.pi * 2.0
    drad = math.sqrt(2.0) / float(GridCls)

    for iv in range(Nviews):

        xv,yv,zv = views[iv]
       
        radv    = math.sqrt(2.0 * (1.0-zv))
        scaleSD = math.sqrt(2.0 / (1.0 + zv))

        if radv > 0 :
           angv = np.arctan2(yv, xv)
           if angv < 0.0 :
              angv = angv + pi2
        else:
           angv = 0.0 
  
        mv1   = int(radv / drad) 
        mv2   = mv1 + 1

#       Masset grid
      
        if GridType == 'mas' :  
           kfull2 = (2 * mv2 - 1) * GridFst
           dphi2  = pi2 / float(kfull2)
           kv     = int(angv / dphi2) + 1       

#       Rosca grid : check outer circle (index 2) and inner circle (index 1)

        elif GridType == 'ros' :
           rad1, rad2 = drad * mv1, drad * mv2
           kfull1     = 8 * mv1
           kfull2     = 8 * mv2
           dphi2      = pi2 / float(kfull2)
           kv         = int(angv / dphi2)
           kv1a, kv1b, kv2a, kv2b = PrevGridNum(kv,mv2)

           if mv1 > 0 :
              dphi1  = pi2 / float(kfull1)
           else :
              dphi1 = 0.0

           ang1a = kv1a * dphi1
           ang1b = kv1b * dphi1
           ang2a = kv2a * dphi2
           ang2b = kv2b * dphi2
           Xd = xv * scaleSD
           Yd = yv * scaleSD
           X1a, Y1a = rad1 * math.cos(ang1a) - Xd , rad1 * math.sin(ang1a) - Yd
           X1b, Y1b = rad1 * math.cos(ang1b) - Xd , rad1 * math.sin(ang1b) - Yd
           X2a, Y2a = rad2 * math.cos(ang2a) - Xd , rad2 * math.sin(ang2a) - Yd
           X2b, Y2b = rad2 * math.cos(ang2b) - Xd , rad2 * math.sin(ang2b) - Yd

#          check if the point is inside the grid (mv,kv) or before / after

           if (X1a * Y2a - X2a * Y1a) < 0.0 : 
              kv = kv2a
           elif (X1b * Y2b - X2b * Y1b) > 0.0 :
              kv = kv2b + 1
           else  :
              kv = kv2b
             
        else :
           print('*** Unknown type of the grid for the views coordinates')
           print('*** Unknown type of the grid for the views coordinates',file=nfmes)
           exit()

        viewsM [iv], viewsK [iv] = mv2  , kv
        viewsR [iv], viewsA [iv] = radv, angv

    views_all = (viewsR,viewsA,viewsM,viewsK,views)

    print('   Sperical grid and views coordinates prepared')

    return views_all                                                                                           
                                                     
#============================
def PrevGridNum(kv,mv) :

#   determine summits of a Rosca cell; 
#   kv1a = kv1b for the diagonal cells a
#   kv2a = m, 3m, 5m, 7m are replaced by the previous values
#   kv2b = m, 3m, 5m, 7m are replaced by the next values

    mv3, mv5, mv7 = mv*3, mv*5, mv*7

    if kv < mv :
       kv1a = kv
    elif kv == mv :
       kv1a = kv - 1
    elif kv < mv3 :
       kv1a = kv - 2
    elif kv == mv3 :
       kv1a = kv - 3
    elif kv < mv5 :
       kv1a = kv - 4
    elif kv == mv5 :
       kv1a = kv - 5
    elif kv < mv7 :
       kv1a = kv - 6
    elif kv == mv7 :
       kv1a = kv - 7
    else :
       kv1a = kv - 8

    kv2a = kv
    kv1b = kv1a + 1
    kv2b = kv2a + 1

    if kv == mv or kv == mv3 or kv == mv5 or kv == mv7 :
       kv2a = kv2a - 1
       kv1b = kv1b - 1

    if kv2b == mv or kv2b == mv3 or kv2b == mv5 or kv2b == mv7 :
       kv2b = kv2b + 1
       kv1b = kv1b - 1

    return kv1a, kv1b, kv2a, kv2b
                                                    
#============================
def CheckViews (ParamGrid,ParamCalc,views_all,CellsXYZ,nfmes):

    GridCls, GridFst, GridType, GridTable = ParamGrid

    if GridType == 'ros' :
       frequencies = CheckViewsRosca(GridCls,ParamCalc,views_all,CellsXYZ,nfmes)

    elif GridType == 'mas' :
       frequencies = CheckViewsMasset(GridCls,GridFst,ParamCalc,views_all,CellsXYZ,nfmes)

    else :
       print('*** Unknown type of the grid for the frequencies calculation')
       print('*** Unknown type of the grid for the frequencies calculation',file=nfmes)
       exit()

    print('   Views frequencies calculated on the spherical grid')

    return frequencies

#============================
def CheckViewsRosca(GridCls,ParamCalc,views_all,CellsXYZ,nfmes):

    AngMax, AngSigma                      = ParamCalc[1],ParamCalc[2]
    ViewFreqType                          = ParamCalc[3]
    viewsR, viewsA, viewsM, viewsK, views = views_all

#   For simplicity, for the Rosca's grid, the contribution is calculated 
#   separately to both cells merged into a single 'sector-type' cell 
#   (their angular indices kg are different) while at the stage of 
#   stats only one of them is treated.

    Nviews = len(views)
    Ncells = float(4.0 * GridCls * GridCls)

    GridCls21 = GridCls * 2 + 1
    GridClsMax = GridCls * 8

    frequencies = [ [0 for j in range(GridClsMax+1)] for i in range(GridCls+1)]

    print('   Analysing views')

#   no blurring; distributioin of grid cells is not important
#   calculate conventional histogram with no surface scaling

    if AngMax == 0.0 :

      for iv in range(Nviews):

        if iv == int(iv/10000) * 10000 : print('         view',iv)

        mv, kv   = viewsM[iv], viewsK[iv]
        frequencies[mv][kv] +=1

#   blurring views contribution to surrounding grid cells
#   with the consecutive normalization of the distribution

    else : 

      dotmin  = math.cos(AngMax)
      delangk  = 2.0 * math.pi / GridClsMax
      kdel     = int(AngMax/delangk) + 1
      delangm   = math.pi / (float(GridCls) * 4.0)
      mdel      = int(AngMax/delangm) + 1
      Mpole     = mdel * 2
      Mcircle   = GridCls - mdel

      for iv in range(Nviews):

        if iv == int(iv/10000) * 10000 : print('         view',iv)

        mv, av   = viewsM[iv], viewsA[iv]
        xv,yv,zv = views [iv]

        convkg   = 4.0 * av / math.pi
        convkgi  = convkg + 4
#                        "+ 4" corresponds to "av + pi" following a half-circle

#       view close to the pole ; check the whole "cup"

        if mv <= mdel:

           for mg in range(1, Mpole+1):
               kfull = mg * 8

               for kg in range(1, kfull + 1):
                  frequencies[mg][kg] += Contribution(xv,yv,zv,CellsXYZ[mg][kg],dotmin,AngSigma)

#       regular view ; check a sperical "rectangle"

        elif mv <= Mcircle:
           m1, m2 = mv - mdel, mv + mdel + 1
           for mg in range(m1,m2):
               kfull = mg * 8

               kmg = int(convkg * mg) + 1 
               k1, k2 = kmg - kdel, kmg + kdel
               if k2-k1 > kfull:
                  k2 = k1 + kfull
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + kfull
                   elif kg > kfull:
                      kgt = kg - kfull
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xv,yv,zv,CellsXYZ[mg][kgt],dotmin,AngSigma)  

#       near-equatorial view ; two opposite views may contribute 

        else:
           m1 = mv - mdel

           for mg in range(m1, GridCls+1):
               kfull = mg * 8

               kmg = int(convkg * mg) + 1 
               k1, k2 = kmg - kdel, kmg + kdel
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + kfull
                   elif kg > kfull:
                      kgt = kg - kfull
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xv,yv,zv,CellsXYZ[mg][kgt],dotmin,AngSigma)

           xvi, yvi, zvi = -xv, -yv, -zv 
           m2  = (GridCls * 2 + 1) - mdel - mv

           for mg in range(m2, GridCls+1):
               kfull = mg * 8

               kmg = int(convkgi * mg) + 1
               k1, k2 = kmg - kdel, kmg + kdel
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + kfull
                   elif kg > kfull:
                      kgt = kg - kfull
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xvi,yvi,zvi,CellsXYZ[mg][kgt],dotmin,AngSigma)   

#     normalize if Gaussian contribution
#     factnorm is Gaussian normalizer, unitsurf is the cell surface

      factnorm  = 1. / (2. * math.pi * AngSigma)
      unitsurf  = 2.0 * math.pi / Ncells
      scalenorm = factnorm * unitsurf

      for mg in range(1,GridCls+1):
          kfull = mg * 8

          for kg in range(1, kfull + 1):
              frequencies[mg][kg] = frequencies[mg][kg] * scalenorm

#     for 'not num' : 
#         frequencies normalized with respect to the 'uniform' value
#         min non-zero value - any positive < 1 (=1 only for uniform)
#         sum of frequencies = Nviews / Ncells 
#     for 'num' :
#         frequencies are count in units
#         min non-zero value is 1
#         sum of frequencies = Nviews

    if ViewFreqType != 'num' :

       scale = Ncells / Nviews

       for mg in range(1,GridCls+1):
           kfull = mg * 8
 
           for kg in range(1, kfull + 1):
               frequencies[mg][kg] = frequencies[mg][kg] * scale

    print('   Rosca grid is used for the frequency calculation')

    return frequencies

#============================
def CheckViewsMasset(GridCls,GridFst,ParamCalc,views_all,CellsXYZ,nfmes):

    AngMax, AngSigma                      = ParamCalc[1],ParamCalc[2]
    ViewFreqType                          = ParamCalc[3]
    viewsR, viewsA, viewsM, viewsK, views = views_all

#   Views frequencies are calculated similarly for both types of the grid.

    Nviews = len(views)
    Ncells = float(GridFst * GridCls * GridCls)

    GridCls21 = GridCls * 2 + 1
    GridClsMax = GridFst * GridCls21

    frequencies = [ [0 for j in range(GridClsMax+1)] for i in range(GridCls+1)]

#   no blurring; distributioin of grid cells is not important
#   calculate conventional histogram with no surface scaling

    if AngMax == 0.0 :

      for iv in range(Nviews):

        if iv == int(iv/10000) * 10000 : print('         view',iv)

        mv, kv   = viewsM[iv], viewsK[iv]
        frequencies[mv][kv] +=1

#   blurring views contribution to surrounding grid cells
#   with the consecutive normalization of the distribution

    else : 

      dotmin  = math.cos(AngMax)
      delangk  = 2.0 * math.pi / GridClsMax
      kdel     = int(AngMax/delangk) + 1
      delangm   = math.pi / (float(GridCls) * 4.0)
      mdel      = int(AngMax/delangm) + 1
      Mpole     = mdel * 2
      Mcircle   = GridCls - mdel

      for iv in range(Nviews):

        mv, av   = viewsM[iv], viewsA[iv]
        xv,yv,zv = views [iv]
        avi      = av + math.pi 

#       view close to the pole ; check the whole "cup"

        if mv <= mdel:

           for mg in range(1, Mpole+1):
               kfull = (2*mg -1) * GridFst

               for kg in range(1, kfull + 1):
                  frequencies[mg][kg] += Contribution(xv,yv,zv,CellsXYZ[mg][kg],dotmin,AngSigma)

#       regular view ; check a sperical "rectangle"

        elif mv <= Mcircle:
           m1, m2 = mv - mdel, mv + mdel + 1
   
           for mg in range(m1,m2):
               kfull = (2*mg -1) * GridFst

               kmg = int(av * kfull / (2.0 * math.pi) ) + 1 
               k1, k2 = kmg - kdel, kmg + kdel
               if k2-k1 > kfull:
                  k2 = k1 + kfull
  
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + kfull
                   elif kg > kfull:
                      kgt = kg - kfull
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xv,yv,zv,CellsXYZ[mg][kgt],dotmin,AngSigma)  

#       near-equatorial view ; two opposite views may contribute 

        else:
           m1 = mv - mdel

           for mg in range(m1, GridCls+1):
               kfull = (2*mg -1) * GridFst

               kmg = int(av * kfull / (2.0 * math.pi) ) + 1 
               k1, k2 = kmg - kdel, kmg + kdel
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + kfull
                   elif kg > kfull:
                      kgt = kg - kfull
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xv,yv,zv,CellsXYZ[mg][kgt],dotmin,AngSigma)

           xvi, yvi, zvi = -xv, -yv, -zv 
           m2  = (GridCls * 2 + 1) - mdel - mv

           for mg in range(m2, GridCls+1):
               kfull = (2*mg -1) * GridFst

               kmg = int(avi * kfull / (2.0 * math.pi) ) + 1 
               k1, k2 = kmg - kdel, kmg + kdel
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + kfull
                   elif kg > kfull:
                      kgt = kg - kfull
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xvi,yvi,zvi,CellsXYZ[mg][kgt],dotmin,AngSigma)   

#     normalize if Gaussian contribution
#     factnorm is Gaussian normalizer, unitsurf is the cell surface

      factnorm  = 1. / (2. * math.pi * AngSigma)
      unitsurf  = 2.0 * math.pi / Ncells
      scalenorm = factnorm * unitsurf

      for mg in range(1,GridCls+1):
          kfull = (2*mg -1) * GridFst

          for kg in range(1, kfull + 1):
              frequencies[mg][kg] = frequencies[mg][kg] * scalenorm

#     for 'not num' : 
#         frequencies normalized with respect to the 'uniform' value
#         min non-zero value - any positive < 1 (=1 only for uniform)
#         sum of frequencies = Nviews / Ncells 
#     for 'num' :
#         frequencies are count in units
#         min non-zero value is 1
#         sum of frequencies = Nviews

    if ViewFreqType != 'num' :

       scale = Ncells / Nviews

       for mg in range(1,GridCls+1):
           kfull = (2*mg -1) * GridFst
 
           for kg in range(1, kfull + 1):
               frequencies[mg][kg] = frequencies[mg][kg] * scale

    print('   Masset grid is used for the frequency calculation')

    return frequencies

#============================
def StatFrequencies (ParamGrid,ParamStat,ParamCalc,CellsFreq,views_all,nfmes):

#   bound values and validation

    GridCls, GridFst, GridType, GridTable      = ParamGrid
    HistNum, HistBottom, HistTop, HistViewStep = ParamStat
    viewsR, viewsA, viewsM, viewsK, views      = views_all
    AngMax, ViewFreqType                       = ParamCalc[1],ParamCalc[3]

    ControlSum, GridSum, SumNul, FreqMax, FreqMin = 0.0, 0.0, 0.0, 0.0, 100.0
    Nviews = float(len(views))

    for mg in range(1,GridCls+1):
        mg3, mg5, mg7 = mg * 3, mg * 5, mg * 7
        if GridType == 'ros' :     
           kfull = mg * 8
        else :
           kfull = (2*mg -1) * GridFst

        for kg in range(1, kfull + 1):
            if GridType != 'ros' or (kg != mg and kg != mg3 and kg != mg5 and kg != mg7) :
               GridSum    = GridSum   + 1
               freq       = CellsFreq[mg][kg]

               ControlSum = ControlSum + freq
               if freq > FreqMax:
                  FreqMax = freq
               if freq <= 0.0:
                  SumNul = SumNul +1
               elif freq < FreqMin:
                  FreqMin = freq

    FreqMinLog , FreqMaxLog  = math.log2(FreqMin) , math.log2(FreqMax)

    if SumNul == GridSum:
        print('*** NO CONTRIBUTION FOUND ***')
        print('*** NO CONTRIBUTION FOUND ***',file=nfmes)

        exit()

    print('',file=nfmes)
    print('============== data statistics ==============',file=nfmes)
    print('',file=nfmes)
    print('Total number of cells, number of empty cells',
                f'{int(GridSum):10}  {int(SumNul):11}',file=nfmes)

#   number of views analyzed

    if ViewFreqType == 'num' :
       print('cell values are counts of the contributing views',file=nfmes)
       if abs(ControlSum / Nviews - 1.0) > 0.1 :
          print('*** sum if frequences is different from expected:',ControlSum,Nviews)
          exit()
    else :
       print('cell values are frequencies normalized wrt uniform value',file=nfmes)
       if abs(ControlSum / GridSum - 1.0) > 0.1 :
          print('*** sum if frequences is different from expected:',ControlSum,GridSum)
    print(' ',file=nfmes)

#   histogram preparation
#   when taking integer part, it is supposed that HistBottom < 0

    if HistBottom == 'unkn': 
       HistBottomInt =  int(FreqMinLog)
       if FreqMinLog < 0.0 : HistBottomInt =  HistBottomInt - 1
       HistBottom    =  float(HistBottomInt)

    if ViewFreqType == 'num' and AngMax == 0.0 :
       HistBottom    =  0.0

    if HistTop    == 'unkn': HistTop    =  FreqMaxLog
    HistTopInt    =  int(HistTop)
    HistTop       =  float(HistTopInt)

    if HistNum    == 'unkn':
       HistNum    =  int(HistTop-HistBottom)     

    print('Number of bins for the Log2 Freq-histogram',f'{HistNum:12}',file=nfmes)
    print('Minimal positive and Maximal cell values',' '*12,
             f'{FreqMin:.5E}  {FreqMax:.5E}',file=nfmes)
    print('Log2 limits for frequencies histogram ',' '*9,
                               f'{HistBottom:8.1f}     {HistTop:8.1f}',file=nfmes)
    
    Hist0    = [0 for i in range(HistNum+2)]
    Hist1    = [0 for i in range(HistNum+2)]
    Hist2    = [0 for i in range(HistNum+2)]
    CumHist1 = [0 for i in range(HistNum+2)]
    CumHist2 = [0 for i in range(HistNum+2)]

#   histograms calculation

#   bin 0 counts number of empty cells
#   bin 1 counts number of cells with frequency slightly larger
#       than HistBottom and eventually with non-zero values smaller
#   bin HistNum+1 counts number of cells wuth frequency >HistTop
 
    StepHist   = float(HistTop - HistBottom) / float(HistNum)
    SumTotal   = 4 * GridCls**2
    bin1, bin2 = 1.0 / float(GridSum) , 1.0 / float(GridSum-SumNul)

    for mg in range(1, GridCls+1):
        mg3, mg5, mg7 = mg * 3, mg * 5, mg * 7
        if GridType == 'ros' :     
           kfull = mg * 8
        else :
           kfull = (2*mg -1) * GridFst

        for kg in range(1, kfull + 1):
            if GridType != 'ros' or (kg != mg and kg != mg3 and kg != mg5 and kg != mg7) :
               freq = CellsFreq[mg][kg]

#              counter ih has "+1" because the first bin counts empty cells
#              and the minimal-value cells are counted in bin 1.

               if freq <= 0.0:
                  ih = 0
               else :
                  freqlog = math.log2(freq)
                  if freqlog <= HistBottom :
                     ih = 1
                  elif freqlog >= HistTop :
                     ih = HistNum+1
                  else:
                     ih = int((freqlog - HistBottom) / StepHist) + 1

               Hist0[ih] += 1

    for ih in range(HistNum+2):
        hist                 = Hist0[ih]
        Hist1[ih], Hist2[ih] = hist * bin1, hist * bin2
    Hist2[0] = 0

#   cumulative distributions

    CumHist1[0], CumHist2[0] = Hist1[0], Hist2[0]

    for ih in range(HistNum+1):
        CumHist1[ih+1] = CumHist1[ih] + Hist1[ih+1]
        CumHist2[ih+1] = CumHist2[ih] + Hist2[ih+1]

#   save statistics

    ParamStatN = (HistNum,HistBottom,HistTop,HistViewStep)
    Hist_all   = (Hist0,Hist1,Hist2,CumHist1,CumHist2)

    print('   Grid frequency statistics calculated')

    StatCells = (ControlSum, GridSum, FreqMin,FreqMax,  \
                 FreqMinLog,FreqMaxLog)

    return CellsFreq,StatCells,Hist_all,ParamStatN

#============================
def Contribution(xv,yv,zv,grid,dotmin,AngSigma):

    dotprod = grid[0]*xv + grid[1]*yv + grid[2]*zv

    freq = math.exp((dotprod-1.) / AngSigma)

#    if dotprod >= dotmin:
#       freq = math.exp((dotprod-1.) / AngSigma)
#    else:
#       freq = 0.0

    return freq
  
#============================
def CellsToViews (ParamGrid,ParamCalc,ParamStat,views_all,GridFreq,nfmes):

    viewsR, viewsA, viewsM, viewsK, views  = views_all
    GridCls, GridFst, GridType, GridTable  = ParamGrid
    ViewFreqType                           = ParamCalc[3]
    HistViewStep                           = ParamStat[3]
    TableType, TableFreqStep, TableFreqRad = GridTable

    Nviews = len(views)

#   view frequency = frequency of the cell

    if ViewFreqType == 'cell' or ViewFreqType == 'num':
       ViewsFreq = ViewsDirect(views_all,GridFreq)

#    option "get frequencies by interpolation from neigboring cells" was removed
#
#    elif ViewFreqType == 'int' :

#   this option is valid only for the GridType = 'ros' (Rosca grid)
#   first get interpolated frequency for the grid nodes
#   then get linear interpolations for the views frequencies

#         NodesFreq = CellsToGridRosca(GridFreq,GridCls,nfmes)
  
#         ViewsFreq = GridToViewsRosca(views_all,NodesFreq,GridCls,nfmes)

    else :
       print('*** wrong Type of the method to calculate Views frequencies',ViewFreqType)
       print('*** wrong Type of the method to calculate Views frequencies',ViewFreqType, \
                                                                            file=nfmes)
       exit()

#   normalize views frequencies
#   by construction, ViewsFreq[iv] are always positive

    SumFreqViews = 0.0
    RmsFreqViews = 0.0
    MinFreqViews = ViewsFreq[0]
    MaxFreqViews = MinFreqViews
    
    for iv in range(Nviews) :
        view_iv       = ViewsFreq[iv]
        ViewsFreq[iv] = view_iv
        SumFreqViews  += view_iv
        RmsFreqViews  += view_iv * view_iv
        if view_iv > MaxFreqViews :
           MaxFreqViews = view_iv
        if view_iv < MinFreqViews :
           MinFreqViews = view_iv

    MeanFreqViews = SumFreqViews / float(Nviews)
    RmsFreqViews  = math.sqrt( RmsFreqViews/float(Nviews) - MeanFreqViews * MeanFreqViews)

#   by construction there are frequencies both < 1 and > 1
#   unless number of views is analyzed

    HistViewMin = int(math.log2(MinFreqViews))
    HistViewMax = int(math.log2(MaxFreqViews))

    if MinFreqViews < 1.0 :
       HistViewMin = HistViewMin -1
    HistViewNum = int( (HistViewMax - HistViewMin) / HistViewStep )

    HistViews = [0 for ih in range(HistViewNum + 1)]

    for iv in range(Nviews) :
        ih = int ( (math.log2(ViewsFreq[iv]*1.0001) - HistViewMin) / HistViewStep )
        if ih > HistViewNum : ih = HistViewNum

        HistViews[ih] += 1

#   analysis of missed / underrepresented regions

    TableViews = []
    NumTables = len(TableFreqRad)

    if TableType == 'num' :
       for irad in range(NumTables) :
           TableRad  = TableFreqRad[irad]
           TableFreq = GetTableFreqN(views,ViewsFreq,TableFreqStep,TableRad,nfmes)
           TableViews.append(TableFreq)

    elif TableType == 'freq' :
       for irad in range(NumTables) :
           TableRad  = TableFreqRad[irad]
           TableFreq = GetTableFreqF(views,ViewsFreq,TableFreqStep,TableRad,nfmes)
           TableViews.append(TableFreq)

    StatFreqViews = MeanFreqViews,MinFreqViews,MaxFreqViews,RmsFreqViews
    StatFreqHist  = HistViewMax, HistViewMin, HistViewNum
    StatViews     = (StatFreqViews,StatFreqHist,HistViews,TableViews)

    print('   Frequencies for individual views calculated')

    return ViewsFreq,StatViews
  
#============================
def GetTableFreqN(views,ViewsFreq,GridStep,Rad,nfmes):

    cutdot = math.cos(Rad * math.pi / 180.)

    NViews  = len(ViewsFreq)
    Ntheta  = int(90.0001 / GridStep)
    Nphi    = int(Ntheta * 4)
    StepRad = math.pi * 0.5 / Ntheta

#   number of views within the spherical segment ('hat') of the angle Rad

    TableFreq = [[0 for iph in range(Ntheta+1)] for ith in range(Nphi)]

    for iph in range(Nphi) :
        phi = StepRad * iph
        phcos = math.cos(phi)
        phsin = math.sin(phi)

        for ith in range(Ntheta+1) :
            theta = StepRad * ith
            thcos = math.cos(theta)
            thsin = math.sin(theta)

            xgrid = thsin * phcos
            ygrid = thsin * phsin
            zgrid = thcos
            kview = 0

            for iview in range (NViews):
               xv, yv, zv = views[iview]
               dot = xgrid * xv + ygrid * yv + zgrid * zv
               if dot < 0.0 : dot = -dot
               if dot > cutdot :
                  kview += 1

            TableFreq[iph][ith] = kview

    return TableFreq
   
#============================
def GetTableFreqF(views,ViewsFreq,GridStep,Rad,nfmes):

    cutdot = math.cos(Rad * math.pi / 180.)

    NViews  = len(ViewsFreq)
    Ntheta  = int(90.0001 / GridStep)
    Nphi    = int(Ntheta * 4)
    StepRad = math.pi * 0.5 / Ntheta

#   mean frequency within the spherical segment ('hat') of the angle Rad
#   (this means : Kview/SegmentSurface : Nviews/DemisphereSurface) 

    surf = (1.0 - math.cos(Rad * math.pi / 180.)) / 2.0
    scale = NViews * surf

    TableFreq = [[0.0 for iph in range(Ntheta+1)] for ith in range(Nphi)]

    for iph in range(Nphi) :
        phi          = StepRad * iph
        phcos, phsin = math.cos(phi) , math.sin(phi)

        for ith in range(Ntheta+1) :
            theta        = StepRad * ith
            thcos, thsin = math.cos(theta) , math.sin(theta)

            xgrid, ygrid, zgrid = thsin * phcos , thsin * phsin , thcos
            kview               = 0

            for iview in range (NViews):
               xv, yv, zv = views[iview]
               dot = xgrid * xv + ygrid * yv + zgrid * zv
               if dot < 0.0 : dot = -dot
               if dot > cutdot :
                  kview += 1

            TableFreq[iph][ith] = float(kview) / scale

    return TableFreq
 
#============================
def ViewsDirect(views_all,GridFreq):

    viewsR, viewsA, viewsM, viewsK, views = views_all
    Nviews = len(viewsM)

    ViewsFreq = [0 for i in range(Nviews)]

    for iv in range (Nviews):
        mv, kv        = viewsM[iv], viewsK[iv]
        ViewsFreq[iv] = GridFreq[mv][kv]

    return ViewsFreq

#============================
def DrawAll (ParamGrid,ParamStat,ParamFig,ParamColor,ParamAux,ParamReper,
             ParamCalc,views_all,Hist_all,CellsFreq,StatCells,ViewsFreq) :

    GridCls, GridFst, GridType, GridTable                    = ParamGrid
    ColorBox, ColorHist, ColorPlus, ColorMap, ColorViews     = ParamColor
    Hist0,Hist1, Hist2, CumHist1, CumHist2                   = Hist_all
    viewsR, viewsA, viewsM, viewsK, views                    = views_all
    ControlSum,GridSum,FreqMin,FreqMax,FreqMinLog,FreqMaxLog = StatCells   
    ViewFreqType                                             = ParamCalc[3]
    Nviews = len(viewsR)

    ColorBox1, ColorBox2, ColorBox4, BoxTitle       = ColorBox
    HistNum, HistBottom, HistTop, HistViewStep      = ParamStat
    ParamDraw, SizeX, ShiftH, ShowDots              = ParamFig
    ParamAuxBar, TitleX, ParamAuxHist, ParamAuxSize = ParamAux
    RepRadius, RepAngle, RepLine, RepPhi0, RepColor = ParamReper

    BarNum, BarBottom, BarTop, BarInt    = ParamAuxBar
    HistBarWidth, HistBarShift           = ParamAuxHist
    SizeLabMin, SizeLabMax, SizeLabScale = ParamAuxSize
    DrawHist, DrawMap, DrawView          = ParamDraw

    FreqLog1 = [0 for i in range(HistNum+2)]
    FreqLog2 = [0 for i in range(HistNum+2)]

#   define internal parameters

    if DrawHist == 'yes' :
       FlagH =  1
    else :
       FlagH =  0

    if DrawMap == 'yes' :
       FlagM =  1
    else :
       FlagM =  0

    if DrawView == 'yes' :
       FlagV =  1
    else :
       FlagV =  0 

    RepRadius4 = float(RepRadius * 4)

#   color scale with respect to the mean frequency in the log2-scale 
#   map from (BarBottom , BarTop) to (0 , BarInt)

    if BarTop == 'unkn' : 
       BarTop = HistTop
    if BarBottom == 'unkn' : 
       BarBottom = HistBottom

    if BarNum == 'unkn' :
       BarNum = int(BarTop - BarBottom) + 1
       if BarNum > 11 : BarNum = 11
       if BarNum <  3 : BarNum = 11

    ColorStep = float(BarInt) / (float(BarTop - BarBottom))

#   colors and the colorbar

    colors    = cm.get_cmap(ColorMap)
    newcolors = colors(np.linspace(0, 1, BarInt+1))

    ColorScale = mpl.colors.Normalize(vmin=BarBottom, vmax=BarTop)
    ColorBar   = plt.cm.ScalarMappable(cmap=colors, norm=ColorScale)
    ColorBar.set_array([])

#   define the label size

    SizeLab = SizeX * 2.54
    if SizeLab < SizeLabMin :
       SizeLab = SizeLabMin
    elif SizeLab > SizeLabMax : 
       SizeLab = SizeLabMax
    LabelSizeA, LabelSizeB = int(SizeLab),              int(SizeLab)
    LabelSizeC             = int(SizeLab)
    LabelSizeT, LabelSizeF = int(SizeLab*SizeLabScale), int(SizeLab*SizeLabScale)

    SizeY   = SizeX
    SizeYH  = SizeY * 0.95
    SizeX2  = 2.0 * SizeX + ShiftH
    SizeY2  = 2.0 * SizeY

#   define the figure composition ; prepare the frames for images

    title_P  = 'VUE 5.2'
    title_H  = 'Histogram of log2(Freq/Aver)'
    title_C  = 'Cumulative func. of log2(Freq/Aver)'
    title_D  = 'Frequency map'
    title_V  = 'Individual views'
    title_B  = 'log2(f)'
    if ViewFreqType == 'num' :
       title_H = 'Histogram of log2(N Views)'
       title_C = 'Cumulative func. of log2(N Views)'
       title_B = 'log2(N)'
    else :
       title_H  = 'Histogram of log2(freq/aver)'
       title_C  = 'Cumulative func. of log2(freq/aver)'
       title_B  = 'log2(f/a)'

    Nmaps   = FlagM + FlagV

    if FlagH + Nmaps <= 0 :
       return

    if FlagH > 0 : 
       if Nmaps == 0 :
          fig, (axe1,axe2) = plt.subplots(1,2,figsize = (SizeX2,SizeYH))
          axe1.set_title(title_H,fontsize=LabelSizeT)
          axe2.set_title(title_C,fontsize=LabelSizeT)

       else :
          fig, ((axe3,axe4),(axe1,axe2)) = plt.subplots(2,2,figsize = (SizeX2,SizeY2))
          axe1.set_title(title_H,fontsize=LabelSizeT)
          axe2.set_title(title_C,fontsize=LabelSizeT)
          axe3.set_title(title_D,fontsize=LabelSizeT)
          axe4.set_title(title_V,fontsize=LabelSizeT)
          axe3.get_xaxis().set_visible(False)
          axe3.get_yaxis().set_visible(False)
          axe4.get_xaxis().set_visible(False)
          axe4.get_yaxis().set_visible(False)
          ColorBarBox = ColorBox4

    else:
       ColorBarBox = ColorBox1
       if Nmaps == 2 :
          fig, (axe3,axe4) = plt.subplots(1,2,figsize = (SizeX2,SizeY))
          axe3.set_title(title_D,fontsize=LabelSizeT)
          axe4.set_title(title_V,fontsize=LabelSizeT)
          axe3.get_xaxis().set_visible(False)
          axe3.get_yaxis().set_visible(False)
          axe4.get_xaxis().set_visible(False)
          axe4.get_yaxis().set_visible(False)
          ColorBarBox = ColorBox2
       elif FlagM == 1 :
          fig, (axe3) = plt.subplots(1,1,figsize = (SizeX,SizeY))
          axe3.set_title(title_D,fontsize=LabelSizeT)
          axe3.get_xaxis().set_visible(False)
          axe3.get_yaxis().set_visible(False)
       else :
          fig, (axe4) = plt.subplots(1,1,figsize = (SizeX,SizeY))
          axe4.set_title(title_V,fontsize=LabelSizeT)
          axe4.get_xaxis().set_visible(False)
          axe4.get_yaxis().set_visible(False)

#   plot colorbar and its title

    if Nmaps > 0 :
       mpl.rcParams['ytick.labelsize'] = LabelSizeB
       CBar = fig.add_axes(ColorBarBox)
       BarTicks=np.linspace(BarBottom, BarTop, BarNum)
       plt.colorbar(ColorBar, cax=CBar, orientation='vertical',ticks=BarTicks)
       CBar.set_yticklabels(['{:4.1f}'.format(i) for i in BarTicks])

       plt.figtext(BoxTitle[0], BoxTitle[1], title_B, fontsize=LabelSizeT)

#   plot main title

    fig.suptitle(title_P,fontsize=LabelSizeF,fontweight='bold',x=TitleX)

#   plot the frequency histogram and the cumulative distribution

    dh = float(HistTop - HistBottom) / float(HistNum)

    for ih in range(HistNum+2) :
        FreqLog1[ih] = dh * (ih-1) + HistBottom
        FreqLog2[ih] = dh * (ih-1) + HistBottom + HistBarShift

    if FlagH > 0 :
       axe1.tick_params(axis='x', labelsize=LabelSizeA)
       axe1.tick_params(axis='y', labelsize=LabelSizeA)
       axe2.tick_params(axis='x', labelsize=LabelSizeA)
       axe2.tick_params(axis='y', labelsize=LabelSizeA)
       axe1.bar(FreqLog1,Hist1,   color=ColorHist,width=HistBarWidth, align='edge')
       axe1.bar(FreqLog2,Hist2,   color=ColorPlus,width=HistBarWidth, align='edge')
       axe2.bar(FreqLog1,CumHist1,color=ColorHist,width=HistBarWidth, align='edge')
       axe2.bar(FreqLog2,CumHist2,color=ColorPlus,width=HistBarWidth, align='edge')

    print('   Histograms ready')

#   plot distribution

    if(FlagM > 0) :

#      draw circle and coarse reper grid

       axe3.add_artist(
                ptc.Circle((0.5,0.5), 0.5, facecolor='white',edgecolor='black', zorder = 1))

       for irad in range (1, RepRadius) :
          angrep = float(irad * math.pi / RepRadius4)
          rad = math.sin(angrep) / math.sqrt(2.0)
          axe3.add_artist(
             ptc.Circle((0.5,0.5), rad, facecolor='None', edgecolor=RepColor, \
                                                              linewidth=RepLine, zorder = 3))

       if RepAngle > 0 :
          radius = ([0.5, 0.5], [1.0, 0.5])
          axe3.add_artist(
             ptc.Polygon(radius, closed=False, color=RepColor, linewidth=RepPhi0, zorder = 3))

       axe3.text(0.90, 0.90,  '45', fontsize=LabelSizeC)
       axe3.text(0.05, 0.90, '135', fontsize=LabelSizeC)
       axe3.text(0.05, 0.05, '225', fontsize=LabelSizeC)
       axe3.text(0.85, 0.05, '315', fontsize=LabelSizeC)

       for iang in range (1, RepAngle) :
          angle  = 2.0 * math.pi * float(iang/RepAngle)
          xr, yr = 0.5 * math.cos(angle) + 0.5, 0.5 * math.sin(angle) + 0.5
          radius = ([0.5, 0.5], [xr, yr])
          axe3.add_artist(
             ptc.Polygon(radius, closed=False, color=RepColor, linewidth=RepLine, zorder = 3))

#      draw map itself

       coord  = [ [0,0] for i in range(4)]

       for m in range(1, GridCls+1) :

#       This we would need to show the equidistant circles on the sphere,
#
#       StepAngle = math.pi / float(4.0 * GridCls)
#           angrep1, angrep2 = float(m-1) *  StepAngle, float(m) *  StepAngle
#           rad1 = math.sin(angrep1) * math.sqrt(2.0)
#           rad2 = math.sin(angrep2) * math.sqrt(2.0)

#        while here we need to show equidistant circles of the uniform grid on disk

           if GridType == 'ros' :
              kfull2  = m * 8
              kfull1 = kfull2 - 8
           else :
              kfull2 = (2*m - 1)  * GridFst
              kfull1 = kfull2

           delang2 = 2.0 * math.pi / float(kfull2) 
           if m > 1 :
              delang1 = 2.0 * math.pi / float(kfull1) 
           else :
              delang1 = 0.0

           rad1 = float(m-1) * 0.5 / float(GridCls)
           rad2 = float(m)   * 0.5 / float(GridCls) 

           coord[0][0], coord[0][1] = rad2 + 0.5 , 0.5
           coord[1][0], coord[1][1] = rad1 + 0.5 , 0.5
           k = 0

           while k < kfull2 :
              coord, k  = GetCellNext(m,k,rad1,rad2,delang1,delang2,coord,GridType)
              freq      = CellsFreq[m][k]
              CellColor = GetColor(freq,BarBottom,BarTop,ColorStep,newcolors,BarInt)

              axe3.add_artist(ptc.Polygon(coord, closed=True, color = CellColor))

    print('   Distribution map ready')

#   draw fine grid
#                  axe3.add_artist(ptc.Polygon(coord, closed=True,
#                                 facecolor='white',edgecolor='black',linewidth=0.5))

#   plot individual views

    if FlagV > 0:

#      draw circle and coarse reper grid

       axe4.add_artist(
               ptc.Circle((0.5,0.5), 0.5, facecolor='white',edgecolor='black', zorder = 1))

       for irad in range (1, RepRadius) :
          angrep = float(irad * math.pi / RepRadius4)
          rad = math.sin(angrep) / math.sqrt(2.0)
          axe4.add_artist(
              ptc.Circle((0.5,0.5), rad, facecolor='None', edgecolor=RepColor, linewidth=RepLine, zorder = 3))

       if RepAngle > 0 :
          radius = ([0.5, 0.5], [1.0, 0.5])
          axe4.add_artist(
             ptc.Polygon(radius, closed=False, color=RepColor, linewidth=RepPhi0, zorder = 3))

       axe4.text(0.90, 0.90,  '45', fontsize=LabelSizeC)
       axe4.text(0.05, 0.90, '135', fontsize=LabelSizeC)
       axe4.text(0.05, 0.05, '225', fontsize=LabelSizeC)
       axe4.text(0.85, 0.05, '315', fontsize=LabelSizeC)

       for iang in range (1, RepAngle) :
          angle  = 2.0 * math.pi * float(iang/RepAngle)
          xr, yr = 0.5 * math.cos(angle) + 0.5, 0.5 * math.sin(angle) + 0.5
          radius = ([0.5, 0.5], [xr, yr])
          axe4.add_artist(
             ptc.Polygon(radius, closed=False, color=RepColor, linewidth=RepLine, zorder = 3))

#      plot view themselves
 
       for iv in range(Nviews):
          rad = viewsR[iv] * 0.5 / math.sqrt(2.0)
          ang = viewsA[iv]
          xv, yv = rad * math.cos(ang) + 0.5 , rad * math.sin(ang) + 0.5

          if   ColorViews == 'top' : freq = BarTop
          elif ColorViews == 'bot' : freq = BarBottom
          else :                     freq = ViewsFreq[iv]

          ViewColor = GetColor(freq,BarBottom,BarTop,ColorStep,
                               newcolors,BarInt)

          axe4.add_artist(ptc.Circle((xv,yv), ShowDots, color = ViewColor, zorder = 2))

    print('   Views map ready')

    plt.show()

    return 
  
#============================
def GetCellNext(m,k0,rad1,rad2,delang1,delang2,coord,GridType) :

    coord[2] = coord[1]
    coord[3] = coord[0]
    k2 = k0 + 1  

    if   GridType == 'mas' :
       k1 = k2

    elif GridType == 'ros' :
       m3, m5, m7 = m*3, m*5, m*7
       if k2 == m or k2 == m3 or k2 == m5 or k2 == m7 :
          k2 += 1

       if k2   < m  :
          k1 = k2
       elif k2 < m3 :
          k1 = k2 - 2
       elif k2 < m5 :
          k1 = k2 - 4
       elif k2 < m7 :
          k1 = k2 - 6
       else :
          k1 = k2 - 8

    angle1 = delang1 * k1
    angle2 = delang2 * k2

    coord[0] = [rad2 * math.cos(angle2) + 0.5 , rad2 * math.sin(angle2) + 0.5 ]
    coord[1] = [rad1 * math.cos(angle1) + 0.5 , rad1 * math.sin(angle1) + 0.5 ]

    return coord, k2

#============================
def GetColor(freq,BarBottom,BarTop,ColorStep,newcolors,BarInt) :

    if freq > 0.0:
       ColorValue  = math.log2(freq)
       if   ColorValue >= BarTop :
            CellColor  =  newcolors[BarInt]
       elif ColorValue <= BarBottom :
            CellColor  =  newcolors[0]
       else:
            CellColor = newcolors[int((ColorValue - BarBottom) * ColorStep)]
    else :
       CellColor = newcolors[0]

    return CellColor

#============================
def PrintStat(ParamGrid,ParamStat,ParamFileN,ParamCalc,Hist_all,
                       CellsFreq,StatCells,ViewsFreq,StatViews,nfmes) :

    GridCls, GridFst, GridType, GridTable                    = ParamGrid
    HistNum,HistBottom,HistTop,HistViewStep                  = ParamStat
    PrintFlag                                                = ParamFileN[0][0]
    AngMax,ViewFreqType                                      = ParamCalc[1],ParamCalc[3]
    Hist0, Hist1, Hist2, CumHist1, CumHist2                  = Hist_all
    ControlSum,GridSum,FreqMin,FreqMax,FreqMinLog,FreqMaxLog = StatCells  
    StatFreqViews,StatFreqHist,HistViews,TableViews          = StatViews
    MeanFreqViews,MinFreqViews,MaxFreqViews,RmsFreqViews     = StatFreqViews
    HistViewMax, HistViewMin, HistViewNum                    = StatFreqHist
    TableType, TableFreqStep, TableFreqRad                   = GridTable

#   print views statistics

    MeanFreqLog = math.log2(MeanFreqViews)
    MinFreqLog  = math.log2(MinFreqViews)
    MaxFreqLog  = math.log2(MaxFreqViews)

    print ('Minimal, Maximal, Mean views frequency  ',9*' ',
          f'{MinFreqViews:10.5f}{MaxFreqViews:12.4f}{MeanFreqViews:10.5f}',file=nfmes)
    print ('Log2 of Min, Max, Mean views frequency  ',9*' ',
          f'{MinFreqLog:10.5f}{MaxFreqLog:12.4f}{MeanFreqLog:10.5f}',file=nfmes)
    print ('Rmsd of views frequency distribution',13*' ',f'{RmsFreqViews:9.4f}',file=nfmes)

    print('',file=nfmes)
    print('Histogram of the views frequencies ',file=nfmes)
    print('   n   Log2(Freq)         Frequency          Nviews   Fraction',file=nfmes)

    nviews = float(len(ViewsFreq))

#    HistViewMin = int(HistViewMin)
#    HistViewMax = int(HistViewMax)

#    if MinFreqViews < 1.0 :
#       HistViewMin = HistViewMin - 1
#    HistViewNum = int ((HistViewMax - HistViewMin)*1.0001 / HistViewStep) + 1

    for ih in range(HistViewNum + 1) :
        HisBotLog = ih * HistViewStep + HistViewMin
        HisTopLog = HisBotLog + HistViewStep
        HisBot    = pow(2.0 , HisBotLog)
        HisTop    = pow(2.0 , HisTopLog)
        HisVal    = HistViews[ih]
        HisRel    = float(HisVal) / nviews
        print(f'{ih:4} {HisBotLog:6.1f}{HisTopLog:6.1f}{HisBot:12.5f}{HisTop:12.5f}',
                       f'{HisVal:9} {HisRel:10.5f}',file=nfmes)

#   print histograms of the grid values 

    print('',file=nfmes)
    print('Histogram of the cell frequencies ',file=nfmes)
    print('   n     Log2(F)          Frequency           Ncells ', 
                             'Hist_All  Hist_Fill  Cumul_All Cumul_Fill', file=nfmes)

    StepHist   = float(HistTop - HistBottom) / float(HistNum)
 
    if PrintFlag > 0:

       Log2Freq2 = HistBottom 
       freq2     = 2.0**Log2Freq2
       print('   0',' '*18,' empty cells     ',
                 f'{Hist0[0]:10}{Hist1[0]:10.5f}',' '*10,f'{CumHist1[0]:10.5f}',file=nfmes)

       for ih in range(1,HistNum+2) :
           Log2Freq1 = HistBottom + (ih-1) * StepHist
           Log2Freq2 = HistBottom +  ih    * StepHist
           freq1     = 2.0**Log2Freq1
           freq2     = 2.0**Log2Freq2
           print(f'{ih:4} {Log2Freq1:6.1f} {Log2Freq2:5.1f} {freq1:11.5f} {freq2:11.5f}',
                 f'{Hist0[ih]:10}{Hist1[ih]:10.5f} {Hist2[ih]:10.5f}',
                 f'{CumHist1[ih]:10.5f} {CumHist2[ih]:10.5f}',file=nfmes)

#   print grid values
 
    if PrintFlag > 1:
       print('',file=nfmes)
       for mg in range(GridCls+1):
           if GridType == 'ros' :
              kfull = mg*8
           else : 
              kfull = (2*mg -1) * GridFst

           for kg in range(kfull + 1):
               if CellsFreq[mg][kg] > 0:
                  print(f'grid cell (GridCls,kgrid)= ({mg:3},{kg:3})',  
                        f'    frequency={CellsFreq[mg][kg]:8.5f}',file=nfmes)

#   print table of views frequencies

    NumTables = len(TableFreqRad)

    for itab in range(NumTables) :

        print('',file=nfmes)
        Table   = TableViews[itab]
        LineTab = Table[0]
        Nphi    = len(Table)
        Ntheta  = len(LineTab)
        StepDeg = 90.0 / (Ntheta-1)
        phi = 0.0

        if TableType == 'num' :
            print('number of views in regions with rad =',  \
                 f'{TableFreqRad[itab]:5.1f} degrees',file=nfmes)
        elif TableType == 'freq' :
            print('average normalized frequency in regions with rad =',  \
                 f'{TableFreqRad[itab]:5.1f} degrees',file=nfmes)

        print('direction to the center of the regions : ',file=nfmes)
        print ('phi = / th =',f'{0.0:3.1f}',   \
                     ''.join(f'{ith*StepDeg:8.1f}' for ith in range(1,Ntheta)),  
                                                                    file=nfmes)
        if TableType == 'num' :
           print(f'{phi:5.1f} ..',   \
              ''.join(f'{LineTab[ith]:8}' for ith in range(Ntheta)),file=nfmes)
        elif TableType == 'freq' :
           print(f'{phi:5.1f} ..',   \
              ''.join(f'{LineTab[ith]:8.3f}' for ith in range(Ntheta)),file=nfmes)

        for iph in range(1,Nphi):
            LineTab = Table[iph]
            phi = iph * StepDeg
            if TableType == 'num' :
               print(f'{phi:5.1f} ..........',   \
                   ''.join(f'{LineTab[ith]:8}' for ith in range(1,Ntheta)),
                                                                       file=nfmes)
            elif TableType == 'freq' :
               print(f'{phi:5.1f} ..........',   \
                   ''.join(f'{LineTab[ith]:8.3f}' for ith in range(1,Ntheta)),
                                                                       file=nfmes)

    print('   Statistics printed')
 
    return

#============================
def GetWeightViews(ParamCalc,StatViews,ViewsFreq,nfmes): 

    ViewWeightType, ViewKappa, ViewCutLog = ParamCalc[4], ParamCalc[5], ParamCalc[6]
    MinFreqViews, MaxFreqViews            = StatViews[0][1], StatViews[0][2] 

    Nviews    = len(ViewsFreq)

    ViewsWeights = [1.0 for iv in range(Nviews)]

#   check parameters values and undate weight mode if necessary

    if ViewWeightType != 'uni' :
       if ViewCutLog <= math.log2(MinFreqViews) :
          ViewCutLog  = math.log2(MinFreqViews)
          if ViewWeightType == 'all' :
             ViewWeightType == 'top'
       if ViewCutLog >= math.log2(MaxFreqViews) :
          ViewWeightType == 'uni'
       else :
          ViewCut = math.pow(2.0, ViewCutLog)

    if ViewWeightType == 'uni' :
       print('   unit Views weights are assigned')

#   calculate weights
#   views frequencies are already normalized to 1

#   reducing over-represented views

    elif ViewWeightType == 'top' :

       SumTop = 0.0
       NumTop = 0
       for iv in range(Nviews) :
           FreqReport = ViewsFreq[iv] / ViewCut
           if FreqReport > 1.0 :
              ViewsWeights[iv] = (ViewKappa * math.log(FreqReport) + 1.0) / FreqReport
              NumTop += 1
              SumTop += ViewsWeights[iv]
       NumCut = int(NumTop - SumTop)

       print('   Views weights calculated')
       print('         Reducing views with the frequency above',ViewCut)
       print('                      i.e. Log2(frequency) above',ViewCutLog)
       print('         Number of views concerned        : ',f'{NumTop:8}')
       print('         Estimated number of views removed: ',f'{NumCut:8}')
       print('         Estimated number of views left   : ',f'{Nviews-NumCut:8}')
       print('',file=nfmes)
       print('Views weights calculated',file=nfmes)
       print('Reducing views with the frequency above  ',f'{ViewCut:12.5f}',file=nfmes)
       print('             i.e. Log2(frequency) above  ',f'{ViewCutLog:12.5f}',file=nfmes)
       print('Number of views concerned            : ',f'{NumTop:8}',file=nfmes)
       print('Estimated number of views removed    : ',f'{NumCut:8}',file=nfmes)
       print('Estimated number of views left       : ',f'{Nviews-NumCut:8}',file=nfmes)

#   reducing over-represented views and expanding under-represented views

    elif ViewWeightType == 'all' :
       SumTop, NumTop = 0.0 , 0
       SumBot, NumBot = 0.0 , 0
       for iv in range(Nviews) :
           FreqReport = ViewsFreq[iv] / ViewCut
           if FreqReport > 1.0 :
              ViewsWeights[iv] = (ViewKappa * math.log(FreqReport) + 1.0) / FreqReport
              NumTop += 1
              SumTop += ViewsWeights[iv]
       NumCut = int(NumTop - SumTop)
       print('   Views weights calculated')
       print('         Reducing views with the frequency above    ',ViewCut)
       print('                        or Log2(frequency) above    ',ViewCutLog)
       print('         Number of views concerned            : ',f'{NumTop:8}')
       print('         Estimated number of views removed    : ',f'{NumCut:8}')
       print('',file=nfmes)
       print('Views weights calculated',file=nfmes)
       print('Reducing views with the frequency above   ',f'{ViewCut:11.5f}',file=nfmes)
       print('                or Log2(frequency) above  ',f'{ViewCutLog:11.5f}',file=nfmes)
       print('Number of views concerned            : ',f'{NumTop:8}',file=nfmes)
       print('Estimated number of views removed    : ',f'{NumCut:8}',file=nfmes)

       FreqReport = ViewCut / MinFreqViews 
       wnorm      = FreqReport / (ViewKappa * math.log(FreqReport) + 1.0) 
       for iv in range(Nviews) :
           if ViewsFreq[iv] / ViewCut <= 1.0 :
              FreqReport = ViewsFreq[iv] / MinFreqViews
              ViewsWeights[iv] = wnorm * (ViewKappa * math.log(FreqReport) + 1.0) / FreqReport
              NumBot           = NumBot + 1
              SumBot           = SumBot + ViewsWeights[iv]
       NumAdd = int(SumBot - NumBot)
       if NumAdd > NumCut :
          Shrink = float(NumCut / NumAdd)
          SumBot = 0.0
          for iv in range(Nviews) :
              if ViewsWeights[iv] > 1.0 :
                 ViewsWeights[iv] = (ViewsWeights[iv] - 1.0) * Shrink + 1.0
                 SumBot           = SumBot + ViewsWeights[iv]     
       NumAdd = int(SumBot - NumBot)
       print('   Views weights calculated - under-represented views completed')
       print('         Number of views concerned            : ',f'{NumBot:8}')
       print('         Estimated number of views to be added: ',f'{NumAdd:8}')
       print('Views weights calculated - under-represented views completed',file=nfmes)
       print('Number of views concerned            : ',f'{NumBot:8}',file=nfmes)
       print('Estimated number of views to be added: ',f'{NumAdd:8}',file=nfmes)
    
    else :
       print('   Views weights calculation fails - weighting scheme not defined ',   \
                 ViewWeightType,file=nfmes)
       print('   Views weights calculation fails - weighting scheme not defined ',   \
                 ViewWeightType)
       exit()

    return ViewsWeights

#============================
def OutViews(ParamFileN,ParamViews,ViewsFreq,ViewsWeights,nfmes): 

    ParamFile, FileStructN                   = ParamFileN
    PrintFlag, FileProj, FileProjW, FileType = ParamFile
    ViewShrink, ViewPert, ViewRand, ViewData = ParamViews

#   check if creating output file with views

    if FileProjW == 'unkn' :
       print('   No output file with views')

    else :

       if FileType == 'test' :
          Nout,Ndel,Nadd,FileProjM = \
               WriteFileTest(ViewsFreq,ViewsWeights,ParamViews,ParamFile,FileStructN,nfmes)

       elif FileType == 'star' :
          Nout,Ndel,Nadd,FileProjM = \
               WriteFileStar(ViewsFreq,ViewsWeights,ParamViews,ParamFile,FileStructN,nfmes)

       elif FileType == 'hed' :
          Nout,Ndel,Nadd,FileProjM = \
               WriteFileHed (ViewsFreq,ViewsWeights,ParamViews,ParamFile,FileStructN,nfmes)

       elif FileProjW != 'unkn':
          print('Stop : unknown file type',FileType)
          print('Stop : unknown file type',FileType,file=nfmes)
          exit()

       Nviews = len(ViewsFreq)

       print(' ',file=nfmes)
       print('============== output views file ==============',file=nfmes)
       print(' ',file=nfmes)

       print('Output file with updated records created; file name ')
       print('     ',FileProjM)
       print('   Number of views in the input file          : ',f'{Nviews:8}')
       print('   Number of views removed                    : ',f'{Ndel:8}')
       print('   Number of views added                      : ',f'{Nadd:8}')
       print('   Number of views in the output file         : ',f'{Nout:8}')

       print('Output file with updated records created; file name ',file=nfmes)
       print('  ',FileProjM,file=nfmes)
       print('Number of views in the input file  : ',f'{Nviews:8}',file=nfmes)
       print('Number of views removed            : ',f'{Ndel:8}',file=nfmes)
       print('Number of views added              : ',f'{Nadd:8}',file=nfmes)
       print('Number of views in the output file : ',f'{Nout:8}',file=nfmes)

    nfmes.close()
 
    return

#============================
def WriteFileTest(ViewsFreq,ViewsWeights,ParamViews,ParamFile,FileStructN,nfmes): 

    PrintFlag, FileProj, FileProjW, FileType = ParamFile
    ViewShrink, ViewPert, ViewRand, ViewData = ParamViews

    nfread  = open(FileProj , 'r')
    nfwrite = open(FileProjW, 'w')

#   check, and correct if necessary, the file-name extension for the known types

    lnfile = len(FileProjW)

    if lnfile < 6 or FileProjW[lnfile-5:lnfile] != '.test' :
       FileProjM = FileProjW + '.test'
    else :
       FileProjM = FileProjW

    if ViewRand != 0 :
       random.seed(ViewRand)

    Nviews           = len(ViewsFreq)
    Nout, Nadd, Ndel = 0, 0, 0

    for iv in range(Nviews) :
        line    = nfread.readline()
        LineCut = line.rstrip('\n')

        freq     = ViewsFreq[iv]
        weight   = ViewsWeights[iv]
        Nrecords = int(weight)
        Vrandom  = weight - Nrecords

        for ic in range(Nrecords) : 
            if random.random() <= ViewShrink :
               WriteLineTest(nfwrite,LineCut,ic,ViewData,ViewPert,freq,weight)
               Nout += 1
               if ic > 0 :
                  Nadd += 1

        if random.random() <= Vrandom and random.random() <= ViewShrink :
            WriteLineTest(nfwrite,LineCut,0,ViewData,ViewPert,freq,weight)
            Nout += 1
            if Nrecords > 0 :
               Nadd +=1

        elif Nrecords == 0 :
            Ndel +=1

    return Nout, Ndel, Nadd, FileProjM

#============================
def WriteLineTest(nfwrite,LineCut,ic,ViewData,ViewPert,freq,weight,nfmes) : 

    LineCont = LineCut.split()

    ang1 = float(LineCont[0])
    ang2 = float(LineCont[1])
    ang3 = float(LineCont[2])
    if ic > 0 :
       ang1p = random.gauss(ang1,ViewPert) 
       ang2p = random.gauss(ang2,ViewPert) 
       ang3p = random.gauss(ang3,ViewPert) 
    else :
       ang1p, ang2p, ang3p = ang1, ang2, ang3

    if ViewData == 'unkn' :
       print(f'{LineCut}',file=nfwrite)

    elif ViewData == 'stat' :
       print(f'{LineCut}{freq:12.7f}{weight:12.7f}',file=nfwrite)

    elif ViewData == 'rep' :
       print(f'{ang1p:12.6f}{ang2p:12.6f}{ang3p:12.6f}',file=nfwrite)

    elif ViewData == 'add' :
       print(f'{LineCut}{ang1p:12.6f}{ang2p:12.6f}{ang3p:12.6f}',file=nfwrite)

    else :
       print('*** unknown type of the output records, ViewData = ',ViewData)
       print('*** unknown type of the output records, ViewData = ',ViewData,file=nfmes)
       exit()

    return

#============================
def WriteFileStar(ViewsFreq,ViewsWeights,ParamViews,ParamFile,FileStructN,nfmes): 

    PrintFlag, FileProj, FileProjW, FileType     = ParamFile
    ViewShrink, ViewPert, ViewRand, ViewData     = ParamViews
    NumTitles,MaxLabel,NumPosAngles,NumLinAngles = FileStructN 

    nfread  = open(FileProj , 'r')
    nfwrite = open(FileProjW, 'w')

#   check, and correct if necessary, the file-name extension for the known types

    lnfile = len(FileProjW)

    if lnfile < 6 or FileProjW[lnfile-5:lnfile] != '.star' :
       FileProjM = FileProjW + '.star'
    else :
       FileProjM = FileProjW

    if ViewRand != 0 :
       random.seed(ViewRand)

    Nviews           = len(ViewsFreq)
    Nout, Nadd, Ndel = 0, 0, 0

#   copy headers including old label descriptions
#   modify label descriptions for the input angles if 'add'
#   add new labels if 'add' or 'stat'

    NumSecAngles = WriteTitleStar(nfread,nfwrite,ViewData,FileStructN,nfmes)

#   output views

    iv = 0 
    
    while True :
       line  = nfread.readline()
       kline = len(line)
       if kline == 0 or line[0] == '#' or line[0] == ';' :
          print(line,file=nfwrite)

       else :
          LineCut = line.rstrip('\n')
          LineCont = line.split()

          freq     = ViewsFreq[iv]
          weight   = ViewsWeights[iv]
          Nrecords = int(weight)
          Vrandom  = weight - Nrecords

          for ic in range(Nrecords) : 
             if random.random() <= ViewShrink :
                WriteLineStar(nfwrite,LineCut,ic,ViewData,NumPosAngles,NumSecAngles,    \
                                      ViewPert,freq,weight,nfmes)
                Nout = Nout +1
                if ic > 0 :
                   Nadd += 1
 
          if random.random() <= Vrandom and random.random() <= ViewShrink :
             WriteLineStar(nfwrite,LineCut,Nrecords,ViewData,NumPosAngles,NumSecAngles,  \
                                      ViewPert,freq,weight,nfmes)
             Nout = Nout +1
             if Nrecords > 0 :
                Nadd +=1

          elif Nrecords == 0 :
             Ndel +=1

          iv += 1
          if iv == Nviews :
             break

#   copy eventual last records

    while True :
       line = nfread.readline()
       if not line :
          break
       else :
          LineCut = line.rstrip('\n')
          print(LineCut,file=nfwrite)

    nfread.close()

    return Nout, Ndel, Nadd, FileProjM

#============================
def WriteTitleStar(nfread,nfwrite,ViewData,FileStructN,nfmes): 

    NumTitles,MaxLabel,NumPosAngles,NumLinAngles = FileStructN 
    NumPosPsi,NumPosRot,NumPosTilt               = NumPosAngles 
    NumLinPsi,NumLinRot,NumLinTilt               = NumLinAngles 

    NumSecAngles = [[0,NumPosPsi], [1,NumPosRot], [2,NumPosTilt]]

    for it in range (NumTitles) :
       line    = nfread.readline()
       LineCut = line.rstrip('\n')

       if ViewData == 'add' :

#         rename all references on angles

          if it == NumLinPsi :
             if NumPosPsi < 9 :
                print(f'_rlnInit_AnglePsi #{NumPosPsi+1:1}',file=nfwrite)
             elif NumPosPsi < 99 :
                print(f'_rlnInit_AnglePsi #{NumPosPsi+1:2}',file=nfwrite)
             else :
                print(f'_rlnInit_AnglePsi #{NumPosPsi+1:3}',file=nfwrite)

          elif it == NumLinRot :
             if NumPosRot < 9 :
                print(f'_rlnInit_AngleRot #{NumPosRot+1:1}',file=nfwrite)
             elif NumPosRot < 99 :
                print(f'_rlnInit_AngleRot #{NumPosRot+1:2}',file=nfwrite)
             else :
                print(f'_rlnInit_AngleRot #{NumPosRot+1:3}',file=nfwrite)

          elif it == NumLinTilt :
             if NumPosTilt < 9 :
                print(f'_rlnInit_AngleTilt #{NumPosTilt+1:1}',file=nfwrite)
             elif NumPosTilt < 99 :
                print(f'_rlnInit_AngleTilt #{NumPosTilt+1:2}',file=nfwrite)
             else :
                print(f'_rlnInit_AngleTilt #{NumPosTilt+1:3}',file=nfwrite)

          else :
             print(LineCut,file=nfwrite)

       else :
          print(LineCut,file=nfwrite)
   
#   add three label descriptions for new angles if required

    if ViewData == 'add' :
       if MaxLabel < 9 :
          print(f'_rlnAnglePsi #{MaxLabel+1:1}',file=nfwrite)
       elif MaxLabel < 99 :
          print(f'_rlnAnglePsi #{MaxLabel+1:2}',file=nfwrite)
       else :
          print(f'_rlnAnglePsi #{MaxLabel+1:3}',file=nfwrite)

       if MaxLabel < 8 :
          print(f'_rlnAngleRot #{MaxLabel+2:1}',file=nfwrite)
       elif MaxLabel < 98 :
          print(f'_rlnAngleRot #{MaxLabel+2:2}',file=nfwrite)
       else :
          print(f'_rlnAngleRot #{MaxLabel+2:3}',file=nfwrite)

       if MaxLabel < 7 :
          print(f'_rlnAngleTilt #{MaxLabel+3:1}',file=nfwrite)
       elif MaxLabel < 97 :
          print(f'_rlnAngleTilt #{MaxLabel+3:2}',file=nfwrite)
       else :
          print(f'_rlnAngleTilt #{MaxLabel+3:3}',file=nfwrite)
   
#   or add two label descriptions for frequencies and weights

    elif ViewData == 'stat' :
       if MaxLabel < 9 :
          print(f'_rlnViewsFrequency #{MaxLabel+1:1}',file=nfwrite)
       elif MaxLabel < 99 :
          print(f'_rlnViewsFrequency #{MaxLabel+1:2}',file=nfwrite)
       else :
          print(f'_rlnViewsFrequency #{MaxLabel+1:3}',file=nfwrite)

       if MaxLabel < 8 :
          print(f'_rlnViewsWeights #{MaxLabel+2:1}',file=nfwrite)
       elif MaxLabel < 98 :
          print(f'_rlnViewsWeights #{MaxLabel+2:2}',file=nfwrite)
       else :
          print(f'_rlnViewsWeights #{MaxLabel+2:3}',file=nfwrite)

#   nothing to do if 'unkn' (structrue of records is kept)

    elif ViewData == 'rep':

       NumSecAngles.sort(key=lambda NumSecAngles: NumSecAngles[1])

    elif ViewData != 'unkn':
        print('*** unknown type of the output records, ViewData = ',ViewData)
        print('*** unknown type of the output records, ViewData = ',ViewData,file=nfmes)
        exit()

    return NumSecAngles

#============================
def WriteLineStar(nfwrite,LineCut,ic,ViewData,NumPosAngles,NumSecAngles,ViewPert,freq,weight,nfmes) : 

    NumPosPsi,NumPosRot,NumPosTilt = NumPosAngles 
    NumSecPsi,NumSecRot,NumSecTilt = NumPosAngles 

    LineCont = LineCut.split()
    ang1 = float(LineCont[NumPosPsi])
    ang2 = float(LineCont[NumPosRot])
    ang3 = float(LineCont[NumPosTilt])
    if ic > 0 :
       ang1p = random.gauss(ang1,ViewPert) 
       ang2p = random.gauss(ang2,ViewPert) 
       ang3p = random.gauss(ang3,ViewPert) 
    else :
       ang1p, ang2p, ang3p = ang1, ang2, ang3

    if ViewData == 'unkn' :
       print(f'{LineCut}',file=nfwrite)

    elif ViewData == 'stat' :
       print(f'{LineCut}{freq:12.7f}{weight:12.7f}',file=nfwrite)

    elif ViewData == 'add' :
       print(f'{LineCut}{ang1p:12.6f}{ang2p:12.6f}{ang3p:12.6f}',file=nfwrite)

    elif ViewData == 'rep' :
       if ic == 0 :
          print(f'{LineCut}',file=nfwrite)
       else :
          array_angles = [ang1p, ang2p, ang3p]
          fin_pos  = elements_position(LineCut)
          Line1 = ReplaceValue(LineCut,NumSecAngles[2][1],fin_pos,array_angles[NumSecAngles[2][0]])
          Line2 = ReplaceValue(Line1  ,NumSecAngles[1][1],fin_pos,array_angles[NumSecAngles[1][0]])
          Line3 = ReplaceValue(Line2  ,NumSecAngles[0][1],fin_pos,array_angles[NumSecAngles[0][0]])
          print(f'{Line3}',file=nfwrite)

    else :
       print('*** unknown type of the output records, ViewData = ',ViewData)
       print('*** unknown type of the output records, ViewData = ',ViewData,file=nfmes)
       exit()

    return

#============================
def ReplaceValue(Line,nelem,fin_pos,ValueNew) :

    LenLine = len(Line)
    if nelem == 0 :
       fin_prev = 0
    else :
       fin_prev = fin_pos[nelem-1]+1
    fin_elem = fin_pos[nelem]

    LenElem   = fin_elem - fin_prev
    if LenElem >= 12 :
       NumDigits = 6
    elif LenElem >= 10 :
       NumDigits = LenElem - 6
    else :
       print('*** something wrong with the format of angles')
       print(Line)
       exit()

    form = f'{LenElem:2}' + '.' + f'{NumDigits:1}' + 'f'
    ValueForm = f'{ValueNew:{form}}'

    LineNew = Line[0:fin_prev+1] + ValueForm + Line[fin_elem+1:LenLine]

    return LineNew

#-------------
def elements_position(Line) :
    fin_pos = []
    LenLine = len(Line)
              
    for ip in range(LenLine-1) :
       if Line[ip] != ' ' and Line[ip+1] == ' ' :
          fin_pos.append(ip)

    fin_pos.append(LenLine)

    return fin_pos

#============================
def WriteFileHed(ViewsFreq,ViewsWeights,ParamViews,ParamFile,FileStructN,nfmes): 

    PrintFlag, FileProj, FileProjW, FileType = ParamFile
    ViewShrink, ViewPert, ViewRand, ViewData = ParamViews

    nfread  = open(FileProj , 'r')
    nfwrite = open(FileProjW, 'w')

#   detailed Imagic format (not required here)
#   form = '@14i 4s 2i 12f 80s f 14i 3f 31i 3f 2i 3f i 13f 2i 77f 228s'

#   format to get Nblocks
    form_Nblocks  = '@12s i 1008s'

    nbytes = 1024

#   check, and correct if necessary, the file-name extension for the known types

    lnfile = len(FileProjW)

    if lnfile < 5 or FileProjW[lnfile-4:lnfile] != '.hed' :
       FileProjM = FileProjW + '.hed'
    else :
       FileProjM = FileProjW

    if ViewRand != 0 :
       random.seed(ViewRand)

    Nviews           = len(ViewsFreq)
    Nout, Nadd, Ndel = 0, 0, 0

    for iv in range(Nviews) :
        line    = nfread.readline()
        LineCut = line.rstrip('\n')

        freq     = ViewsFreq[iv]
        weight   = ViewsWeights[iv]
        Nrecords = int(weight)
        Vrandom  = weight - Nrecords

        for ic in range(Nrecords) : 
            if random.random() <= ViewShrink :
               WriteLineHed(nfwrite,LineCut,ic,ViewData,ViewPert,freq,weight)
               Nout = Nout +1
               if ic > 0 :
                  Nadd += 1

        if random.random() <= Vrandom and random.random() <= ViewShrink :
            WriteLineHed(nfwrite,LineCut,0,ViewData,ViewPert,freq,weight)
            Nout = Nout +1
            if Nrecords > 0 :
               Nadd +=1

        elif Nrecords == 0 :
            Ndel +=1

#       eventual Nblocks following the record

        LineCont = struct.unpack(form_Nblocks, LineCut)
        Nblocks = LineCont[1]

        if Nblocks > 1 :
           for ib in range(Nblocks-1) :
               line = nfproj.read(nbytes)
               nfwrite.write(line)

    return Nout, Ndel, Nadd, FileProjM

#============================
def WriteLineHed(nfwrite,LineCut,ic,ViewData,ViewPert,freq,weight,nfmes) : 

#   detailed Imagic format (not required here)
#   form = '@14i 4s 2i 12f 80s f 14i 3f 31i 3f 2i 3f i 13f 2i 77f 228s'


#   format to extract the number of blocks, angles and space for frequency and weight

    form_in  = '@12s i 232s 4s 4s 3f 756s'
    form_out = '@12s i 232s  f  f 3f 756s'

    LineCont = struct.unpack(form_in, LineCut)

    ang1 = float(LineCont[5])
    ang2 = float(LineCont[6])
    ang3 = float(LineCont[7])
    if ic > 0 :
       ang1p = random.gauss(ang1,ViewPert) 
       ang2p = random.gauss(ang2,ViewPert) 
       ang3p = random.gauss(ang3,ViewPert) 
    else :
       ang1p, ang2p, ang3p = ang1, ang2, ang3

    if ViewData == 'unkn' :
       line_out = LineCut

    if ViewData == 'stat' :
       line_out = struct.pack(form_out,LineCont[0],LineCont[1],LineCont[2],freq,weight,
                                       LineCont[5],LineCont[6],LineCont[7],LineCont[8])

    elif ViewData == 'rep':
       line_out = struct.pack(form_in,LineCont[0],LineCont[1],LineCont[2],LineCont[3],
                                       LineCont[4],ang1p,ang2p,ang3p,LineCont[8])

#   'rep' option is used instead 
    elif ViewData == 'add' :
       line_out = struct.pack(form_in,LineCont[0],LineCont[1],LineCont[2],LineCont[3],
                                       LineCont[4],ang1p,ang2p,ang3p,LineCont[8])
 
    else :
       print('*** unknown type of the output records, ViewData = ',ViewData)
       print('*** unknown type of the output records, ViewData = ',ViewData,file=nfmes)
       exit()

    nfwrite.write(line_out)

    return

####################################################
#
#  MAIN PROGRAM

filemes  = 'VUE.log'
filedata = 'VUE.dat'

#==================== input information =====================

ParamGrid, ParamCalc, ParamStat, ParamFig, ParamColor, ParamFile, ParamAux,   \
           ParamReper, ParamShow, ParamViews,nfmes = InputData(filedata,filemes)

print('Input projections and converting into views...')

views_inp,ParamFileN = GetViews(ParamFile,nfmes)

views = SymViews(views_inp,ParamShow,nfmes)

views = ViewsToShow(views,ParamShow)

#=============== process views and calculate frequencies ==========

print('Preparating spherical grid and views parameters...')

CellsXYZ  = GridCenters(ParamGrid,nfmes)

views_all = ViewsParameters(ParamGrid,views,nfmes)


print('Calculating frequencies and weights... Wait a moment ...')

CellsFreq = CheckViews(ParamGrid,ParamCalc,views_all,CellsXYZ,nfmes)

CellsFreq, StatCells, Hist_all, ParamStatN                                  \
        = StatFrequencies(ParamGrid,ParamStat,ParamCalc,CellsFreq,views_all,nfmes)

ViewsFreq, StatViews = CellsToViews(ParamGrid,ParamCalc,ParamStat,views_all, \
                                                                  CellsFreq,nfmes)

ViewsWeights = GetWeightViews(ParamCalc,StatViews,ViewsFreq,nfmes)

#=============== output results ======================================

print('Writing obtained information...')

PrintStat(ParamGrid,ParamStatN,ParamFileN,ParamCalc,Hist_all,
                                CellsFreq,StatCells,ViewsFreq,StatViews,nfmes)

OutViews(ParamFileN,ParamViews,ViewsFreq,ViewsWeights,nfmes)

print('Preparing diagrams if required... Wait a moment ...')

DrawAll(ParamGrid,ParamStatN,ParamFig,ParamColor,ParamAux,ParamReper,ParamCalc,
                          views_all,Hist_all,CellsFreq,StatCells,ViewsFreq)

print('   Diagrams shown. Done')
